# Urwid unicode character processing tables
#    Copyright (C) 2004-2011  Ian Ward
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Urwid web site: https://urwid.org/


from __future__ import annotations

import re
import typing
import warnings

import wcwidth

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

SAFE_ASCII_RE = re.compile(r"^[ -~]*$")
SAFE_ASCII_BYTES_RE = re.compile(rb"^[ -~]*$")

_byte_encoding: Literal["utf8", "narrow", "wide"] = "narrow"


def get_char_width(char: str) -> Literal[0, 1, 2]:
    """
    Return the screen column width for a single character.

    .. deprecated:: 3.0.4
    """
    warnings.warn(
        "get_char_width is deprecated in favor of wcwidth.width",
        DeprecationWarning,
        stacklevel=2,
    )
    if (width := wcwidth.wcwidth(char)) >= 0:
        return typing.cast("Literal[0, 1, 2]", width)

    return 0


def get_width(o: int) -> Literal[0, 1, 2]:
    """
    Return the screen column width for unicode ordinal o.

    .. deprecated:: 3.0.4
    """
    warnings.warn(
        "get_width is deprecated in favor of wcwidth.width",
        DeprecationWarning,
        stacklevel=2,
    )
    if (width := wcwidth.wcwidth(chr(o))) >= 0:
        return typing.cast("Literal[0, 1, 2]", width)

    return 0


def _decode_grapheme_at(text: bytes, start: int, end: int) -> tuple[str, int]:
    """
    Decode bytes starting at `start` to get the first grapheme cluster.

    :param text: UTF-8 encoded bytes
    :param start: starting byte position
    :param end: ending byte position
    :returns: (grapheme_string, next_byte_position)

    Assumes caller provides valid UTF-8 byte boundaries.
    """
    decoded = text[start:end].decode("utf-8")
    grapheme = next(wcwidth.iter_graphemes(decoded), "")
    grapheme_bytes = grapheme.encode("utf-8")
    return grapheme, start + len(grapheme_bytes)


def decode_one(text: bytes | str, pos: int) -> tuple[int, int]:
    """
    Return (ordinal at pos, next position) for UTF-8 encoded text.
    """
    lt = len(text) - pos

    b2 = 0  # Fallback, not changing anything
    b3 = 0  # Fallback, not changing anything
    b4 = 0  # Fallback, not changing anything

    try:
        if isinstance(text, str):
            b1 = ord(text[pos])
            if lt > 1:
                b2 = ord(text[pos + 1])
            if lt > 2:
                b3 = ord(text[pos + 2])
            if lt > 3:
                b4 = ord(text[pos + 3])
        else:
            b1 = text[pos]
            if lt > 1:
                b2 = text[pos + 1]
            if lt > 2:
                b3 = text[pos + 2]
            if lt > 3:
                b4 = text[pos + 3]
    except Exception as e:
        raise ValueError(f"{e}: text={text!r}, pos={pos!r}, lt={lt!r}").with_traceback(e.__traceback__) from e

    if not b1 & 0x80:
        return b1, pos + 1
    error = ord("?"), pos + 1

    if lt < 2:
        return error
    if b1 & 0xE0 == 0xC0:
        if b2 & 0xC0 != 0x80:
            return error
        if (o := ((b1 & 0x1F) << 6) | (b2 & 0x3F)) >= 0x80:
            return o, pos + 2
        return error
    if lt < 3:
        return error
    if b1 & 0xF0 == 0xE0:
        if b2 & 0xC0 != 0x80:
            return error
        if b3 & 0xC0 != 0x80:
            return error
        if (o := ((b1 & 0x0F) << 12) | ((b2 & 0x3F) << 6) | (b3 & 0x3F)) >= 0x800:
            return o, pos + 3
        return error
    if lt < 4:
        return error
    if b1 & 0xF8 == 0xF0:
        if b2 & 0xC0 != 0x80:
            return error
        if b3 & 0xC0 != 0x80:
            return error
        if b4 & 0xC0 != 0x80:
            return error
        if (o := ((b1 & 0x07) << 18) | ((b2 & 0x3F) << 12) | ((b3 & 0x3F) << 6) | (b4 & 0x3F)) >= 0x10000:
            return o, pos + 4
        return error
    return error


def decode_one_uni(text: str, i: int) -> tuple[int, int]:
    """
    decode_one implementation for unicode strings
    """
    return ord(text[i]), i + 1


def decode_one_right(text: bytes, pos: int) -> tuple[int, int] | None:
    """
    Return (ordinal at pos, next position) for UTF-8 encoded text.
    pos is assumed to be on the trailing byte of a utf-8 sequence.
    """
    if not isinstance(text, bytes):
        raise TypeError(text)
    error = ord("?"), pos - 1
    p = pos
    while p >= 0:
        if text[p] & 0xC0 != 0x80:
            o, _next_pos = decode_one(text, p)
            return o, p - 1
        p -= 1
        if p == p - 4:
            return error
    return None


def set_byte_encoding(enc: Literal["utf8", "narrow", "wide"]) -> None:
    if enc not in {"utf8", "narrow", "wide"}:
        raise ValueError(enc)
    global _byte_encoding  # noqa: PLW0603  # pylint: disable=global-statement
    _byte_encoding = enc


def get_byte_encoding() -> Literal["utf8", "narrow", "wide"]:
    return _byte_encoding


def calc_string_text_pos(text: str, start_offs: int, end_offs: int, pref_col: int) -> tuple[int, int]:
    """
    Calculate the closest position to the screen column pref_col in text
    where start_offs is the offset into text assumed to be screen column 0
    and end_offs is the end of the range to search.

    Iterates by grapheme clusters for emoji ZWJ sequences, flags,
    combining characters, and other multi-codepoint unicode sequences.

    :param text: string
    :param start_offs: starting text position
    :param end_offs: ending text position
    :param pref_col: target column
    :returns: (position, actual_col)
    """
    if start_offs > end_offs:
        raise ValueError((start_offs, end_offs))

    cols = 0
    pos = start_offs
    for grapheme in wcwidth.iter_graphemes(text[start_offs:end_offs]):
        grapheme_width = wcwidth.width(grapheme, control_codes="ignore")
        if grapheme_width + cols > pref_col:
            return pos, cols
        cols += grapheme_width
        pos += len(grapheme)

    return end_offs, cols


def calc_text_pos(text: str | bytes, start_offs: int, end_offs: int, pref_col: int) -> tuple[int, int]:
    """
    Calculate the closest position to the screen column pref_col in text
    where start_offs is the offset into text assumed to be screen column 0
    and end_offs is the end of the range to search.

    text may be unicode or a byte string in the target _byte_encoding

    Returns (position, actual_col).
    """
    if start_offs > end_offs:
        raise ValueError((start_offs, end_offs))

    if isinstance(text, str):
        return calc_string_text_pos(text, start_offs, end_offs, pref_col)

    if not isinstance(text, bytes):
        raise TypeError(text)

    if _byte_encoding == "utf8":
        decoded = text[start_offs:end_offs].decode("utf-8")
        str_pos, cols = calc_string_text_pos(decoded, 0, len(decoded), pref_col)
        byte_offset = len(decoded[:str_pos].encode("utf-8"))
        return start_offs + byte_offset, cols

    # "wide" and "narrow"
    i = start_offs + pref_col
    if i >= end_offs:
        return end_offs, end_offs - start_offs
    if _byte_encoding == "wide" and within_double_byte(text, start_offs, i) == 2:
        i -= 1
    return i, i - start_offs


def calc_width(text: str | bytes, start_offs: int, end_offs: int) -> int:
    """
    Return the screen column width of text between start_offs and end_offs.

    text may be unicode or a byte string in the target _byte_encoding

    Some characters are wide (take two columns) and others affect the
    previous character (take zero columns), while others are grouped
    in sequence by "grapheme boundaries" (Emoji, Skin tones, flags, etc).
    """

    if start_offs > end_offs:
        msg = f"{start_offs=} > {end_offs=}"
        raise ValueError(msg)

    if isinstance(text, str):
        return wcwidth.width(text[start_offs:end_offs], control_codes="ignore")

    if _byte_encoding == "utf8":
        try:
            return wcwidth.width(text[start_offs:end_offs].decode("utf-8"), control_codes="ignore")
        except UnicodeDecodeError as exc:
            warnings.warn(
                "`calc_width` with text encoded to bytes can produce incorrect results"
                f"due to possible offset in the middle of character: {exc}",
                UnicodeWarning,
                stacklevel=2,
            )

        i = start_offs
        sc = 0
        while i < end_offs:
            o, i = decode_one(text, i)
            if (w := wcwidth.wcwidth(chr(o))) > 0:
                sc += w
        return sc
    # "wide", "narrow" or all printable ASCII, just return the character count
    return end_offs - start_offs


def is_wide_char(text: str | bytes, offs: int) -> bool:
    """
    Test if the grapheme cluster at offs within text is wide (2 columns).

    For Unicode strings, extracts the full grapheme cluster starting at offs
    and checks if it renders as wide. This correctly handles multi-codepoint
    graphemes like emoji ZWJ sequences and flags.

    text may be unicode or a byte string in the target _byte_encoding
    """
    if isinstance(text, str):
        grapheme = next(wcwidth.iter_graphemes(text[offs:]))
        return wcwidth.width(grapheme, control_codes="ignore") == 2
    if not isinstance(text, bytes):
        raise TypeError(text)
    if _byte_encoding == "utf8":
        grapheme, _ = _decode_grapheme_at(text, offs, len(text))
        return wcwidth.width(grapheme, control_codes="ignore") == 2
    if _byte_encoding == "wide":
        return within_double_byte(text, offs, offs) == 1
    return False


def move_prev_char(text: str | bytes, start_offs: int, end_offs: int) -> int:
    """
    Return the position of the grapheme cluster before end_offs.

    For Unicode strings, handle multi-codepoint, "grapheme clusters",
    to better measure emoji ZWJ, flags, combining characters, skin tones.
    """
    if start_offs >= end_offs:
        raise ValueError((start_offs, end_offs))
    if isinstance(text, str):
        return wcwidth.grapheme_boundary_before(text, end_offs)
    if not isinstance(text, bytes):
        raise TypeError(text)
    if _byte_encoding == "utf8":
        decoded = text[start_offs:end_offs].decode("utf-8")
        str_pos = len(decoded)
        prev_str_pos = wcwidth.grapheme_boundary_before(decoded, str_pos)
        prefix = decoded[:prev_str_pos]
        return start_offs + len(prefix.encode("utf-8"))
    if _byte_encoding == "wide" and within_double_byte(text, start_offs, end_offs - 1) == 2:
        return end_offs - 2
    return end_offs - 1


def move_next_char(text: str | bytes, start_offs: int, end_offs: int) -> int:
    """
    Return the position of the next grapheme cluster after start_offs.

    For Unicode strings, handle multi-codepoint, "grapheme clusters",
    to better measure emoji ZWJ, flags, combining characters, skin tones.
    """
    if start_offs >= end_offs:
        raise ValueError((start_offs, end_offs))
    if isinstance(text, str):
        grapheme = next(wcwidth.iter_graphemes(text[start_offs:end_offs]))
        return start_offs + len(grapheme)
    if not isinstance(text, bytes):
        raise TypeError(text)
    if _byte_encoding == "utf8":
        _, next_pos = _decode_grapheme_at(text, start_offs, end_offs)
        return next_pos
    if _byte_encoding == "wide" and within_double_byte(text, start_offs, start_offs) == 1:
        return start_offs + 2
    return start_offs + 1


def within_double_byte(text: bytes, line_start: int, pos: int) -> Literal[0, 1, 2]:
    """Return whether pos is within a double-byte encoded character.

    text -- byte string in question
    line_start -- offset of beginning of line (< pos)
    pos -- offset in question

    Return values:
    0 -- not within dbe char, or double_byte_encoding == False
    1 -- pos is on the 1st half of a dbe char
    2 -- pos is on the 2nd half of a dbe char
    """
    if not isinstance(text, bytes):
        raise TypeError(text)
    v = text[pos]

    if 0x40 <= v < 0x7F:
        # might be second half of big5, uhc or gbk encoding
        if pos == line_start:
            return 0

        if text[pos - 1] >= 0x81 and within_double_byte(text, line_start, pos - 1) == 1:
            return 2
        return 0

    if v < 0x80:
        return 0

    i = pos - 1
    while i >= line_start:
        if text[i] < 0x80:
            break
        i -= 1

    if (pos - i) & 1:
        return 1
    return 2



from __future__ import annotations

import unittest

import urwid
from tests.util import SelectableText


class ListBoxCalculateVisibleTest(unittest.TestCase):
    def cvtest(
        self,
        desc: str,
        body,
        focus: int,
        offset_rows: int,
        inset_fraction,
        exp_offset_inset: int,
        exp_cur: tuple[int, int] | None,
    ) -> None:
        with self.subTest(desc):
            lbox = urwid.ListBox(urwid.SimpleListWalker(body))
            lbox.body.set_focus(focus)
            lbox.offset_rows = offset_rows
            lbox.inset_fraction = inset_fraction

            middle, top, bottom = lbox.calculate_visible((4, 5), focus=1)
            offset_inset, focus_widget, focus_pos, _ign, cursor = middle

            if cursor is not None:
                x, y = cursor
                y += offset_inset
                cursor = x, y

            self.assertEqual(exp_offset_inset, offset_inset)
            self.assertEqual(exp_cur, cursor)

    def test1_simple(self):
        T = urwid.Text

        l = [T(""), T(""), T("\n"), T("\n\n"), T("\n"), T(""), T("")]

        self.cvtest(
            "simple top position",
            l,
            3,
            0,
            (0, 1),
            0,
            None,
        )

        self.cvtest(
            "simple middle position",
            l,
            3,
            1,
            (0, 1),
            1,
            None,
        )

        self.cvtest(
            "simple bottom position",
            l,
            3,
            2,
            (0, 1),
            2,
            None,
        )

        self.cvtest(
            "straddle top edge",
            l,
            3,
            0,
            (1, 2),
            -1,
            None,
        )

        self.cvtest(
            "straddle bottom edge",
            l,
            3,
            4,
            (0, 1),
            4,
            None,
        )

        self.cvtest(
            "off bottom edge",
            l,
            3,
            5,
            (0, 1),
            4,
            None,
        )

        self.cvtest(
            "way off bottom edge",
            l,
            3,
            100,
            (0, 1),
            4,
            None,
        )

        self.cvtest(
            "gap at top",
            l,
            0,
            2,
            (0, 1),
            0,
            None,
        )

        self.cvtest(
            "gap at top and off bottom edge",
            l,
            2,
            5,
            (0, 1),
            2,
            None,
        )

        self.cvtest(
            "gap at bottom",
            l,
            6,
            1,
            (0, 1),
            4,
            None,
        )

        self.cvtest(
            "gap at bottom and straddling top edge",
            l,
            4,
            0,
            (1, 2),
            1,
            None,
        )

        self.cvtest(
            "gap at bottom cannot completely fill",
            [T(""), T(""), T("")],
            1,
            0,
            (0, 1),
            1,
            None,
        )

        self.cvtest(
            "gap at top and bottom",
            [T(""), T(""), T("")],
            1,
            2,
            (0, 1),
            1,
            None,
        )

    def test2_cursor(self):
        T, E = urwid.Text, urwid.Edit

        l1 = [T(""), T(""), T("\n"), E("", "\n\nX"), T("\n"), T(""), T("")]
        l2 = [T(""), T(""), T("\n"), E("", "YY\n\n"), T("\n"), T(""), T("")]

        l2[3].set_edit_pos(2)

        self.cvtest(
            "plain cursor in view",
            l1,
            3,
            1,
            (0, 1),
            1,
            (1, 3),
        )

        self.cvtest(
            "cursor off top",
            l2,
            3,
            0,
            (1, 3),
            0,
            (2, 0),
        )

        self.cvtest(
            "cursor further off top",
            l2,
            3,
            0,
            (2, 3),
            0,
            (2, 0),
        )

        self.cvtest(
            "cursor off bottom",
            l1,
            3,
            3,
            (0, 1),
            2,
            (1, 4),
        )

        self.cvtest(
            "cursor way off bottom",
            l1,
            3,
            100,
            (0, 1),
            2,
            (1, 4),
        )

    def test_sized(self):
        lbox = urwid.ListBox(urwid.SimpleListWalker([urwid.Text(str(num)) for num in range(5)]))
        self.assertEqual(5, len(lbox))

    def test_not_sized(self):
        class TestWalker(urwid.ListWalker):
            @property
            def contents(self):
                return self

            @staticmethod
            def next_position(position: int) -> tuple[urwid.Text, int]:
                return urwid.Text(str(position)), position

            @staticmethod
            def prev_position(position: int) -> tuple[urwid.Text, int]:
                return urwid.Text(str(position)), position

        lbox = urwid.ListBox(TestWalker())
        with self.assertRaises(AttributeError) as exc:
            len(lbox)

        self.assertEqual(f"{TestWalker.__name__} is not Sized", str(exc.exception))


class ListBoxChangeFocusTest(unittest.TestCase):
    def cftest(
        self,
        desc: str,
        body,
        pos: int,
        offset_inset: int,
        coming_from,
        cursor: tuple[int, int] | None,
        snap_rows,
        exp_offset_rows: int,
        exp_inset_fraction,
        exp_cur: tuple[int, int] | None,
    ):
        with self.subTest(desc):
            lbox = urwid.ListBox(urwid.SimpleListWalker(body))

            lbox.change_focus((4, 5), pos, offset_inset, coming_from, cursor, snap_rows)

            exp = exp_offset_rows, exp_inset_fraction
            act = lbox.offset_rows, lbox.inset_fraction

            cursor = None
            focus_widget, focus_pos = lbox.body.get_focus()
            if focus_widget.selectable():
                if hasattr(focus_widget, "get_cursor_coords"):
                    cursor = focus_widget.get_cursor_coords((4,))

            self.assertEqual(exp, act)
            self.assertEqual(exp_cur, cursor)

    def test1unselectable(self):
        T = urwid.Text
        l = [T("\n"), T("\n\n"), T("\n\n"), T("\n\n"), T("\n")]

        self.cftest(
            "simple unselectable",
            l,
            2,
            0,
            None,
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "unselectable",
            l,
            2,
            1,
            None,
            None,
            None,
            1,
            (0, 1),
            None,
        )

        self.cftest(
            "unselectable off top",
            l,
            2,
            -2,
            None,
            None,
            None,
            0,
            (2, 3),
            None,
        )

        self.cftest(
            "unselectable off bottom",
            l,
            3,
            2,
            None,
            None,
            None,
            2,
            (0, 1),
            None,
        )

    def test2selectable(self):
        T, S = urwid.Text, SelectableText
        l = [T("\n"), T("\n\n"), S("\n\n"), T("\n\n"), T("\n")]

        self.cftest(
            "simple selectable",
            l,
            2,
            0,
            None,
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable",
            l,
            2,
            1,
            None,
            None,
            None,
            1,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable at top",
            l,
            2,
            0,
            "below",
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable at bottom",
            l,
            2,
            2,
            "above",
            None,
            None,
            2,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable off top snap",
            l,
            2,
            -1,
            "below",
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable off bottom snap",
            l,
            2,
            3,
            "above",
            None,
            None,
            2,
            (0, 1),
            None,
        )

        self.cftest(
            "selectable off top no snap",
            l,
            2,
            -1,
            "above",
            None,
            None,
            0,
            (1, 3),
            None,
        )

        self.cftest(
            "selectable off bottom no snap",
            l,
            2,
            3,
            "below",
            None,
            None,
            3,
            (0, 1),
            None,
        )

    def test3large_selectable(self):
        T, S = urwid.Text, SelectableText
        l = [T("\n"), S("\n\n\n\n\n\n"), T("\n")]
        self.cftest(
            "large selectable no snap",
            l,
            1,
            -1,
            None,
            None,
            None,
            0,
            (1, 7),
            None,
        )

        self.cftest(
            "large selectable snap up",
            l,
            1,
            -2,
            "below",
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "large selectable snap up2",
            l,
            1,
            -2,
            "below",
            None,
            2,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "large selectable almost snap up",
            l,
            1,
            -2,
            "below",
            None,
            1,
            0,
            (2, 7),
            None,
        )

        self.cftest(
            "large selectable snap down",
            l,
            1,
            0,
            "above",
            None,
            None,
            0,
            (2, 7),
            None,
        )

        self.cftest(
            "large selectable snap down2",
            l,
            1,
            0,
            "above",
            None,
            2,
            0,
            (2, 7),
            None,
        )

        self.cftest(
            "large selectable almost snap down",
            l,
            1,
            0,
            "above",
            None,
            1,
            0,
            (0, 1),
            None,
        )

        m = [T("\n\n\n\n"), S("\n\n\n\n\n"), T("\n\n\n\n")]
        self.cftest(
            "large selectable outside view down",
            m,
            1,
            4,
            "above",
            None,
            None,
            0,
            (0, 1),
            None,
        )

        self.cftest(
            "large selectable outside view up",
            m,
            1,
            -5,
            "below",
            None,
            None,
            0,
            (1, 6),
            None,
        )

    def test4cursor(self):
        T, E = urwid.Text, urwid.Edit
        # ...

    def test5set_focus_valign(self):
        T, E = urwid.Text, urwid.Edit
        lbox = urwid.ListBox(urwid.SimpleFocusListWalker([T(""), T("")]))
        lbox.set_focus_valign("middle")
        # TODO: actually test the result


class ListBoxRenderTest(unittest.TestCase):
    def ltest(
        self,
        desc: str,
        body,
        focus: int,
        offset_inset_rows: int,
        exp_text: list[bytes],
        exp_cur: tuple[int, int] | None,
    ) -> None:
        with self.subTest(desc):
            exp_text = [t.encode("iso8859-1") for t in exp_text]
            lbox = urwid.ListBox(urwid.SimpleListWalker(body))
            lbox.body.set_focus(focus)
            lbox.shift_focus((4, 10), offset_inset_rows)
            canvas = lbox.render((4, 5), focus=1)

            self.assertEqual(exp_text, canvas.text)
            self.assertEqual(exp_cur, canvas.cursor)

    def test1_simple(self):
        T = urwid.Text

        self.ltest(
            "simple one text item render",
            [T("1\n2")],
            0,
            0,
            ["1   ", "2   ", "    ", "    ", "    "],
            None,
        )

        self.ltest(
            "simple multi text item render off bottom",
            [T("1"), T("2"), T("3\n4"), T("5"), T("6")],
            2,
            2,
            ["1   ", "2   ", "3   ", "4   ", "5   "],
            None,
        )

        self.ltest(
            "simple multi text item render off top",
            [T("1"), T("2"), T("3\n4"), T("5"), T("6")],
            2,
            1,
            ["2   ", "3   ", "4   ", "5   ", "6   "],
            None,
        )

    def test2_trim(self):
        T = urwid.Text

        self.ltest(
            "trim unfocused bottom",
            [T("1\n2"), T("3\n4"), T("5\n6")],
            1,
            2,
            ["1   ", "2   ", "3   ", "4   ", "5   "],
            None,
        )

        self.ltest(
            "trim unfocused top",
            [T("1\n2"), T("3\n4"), T("5\n6")],
            1,
            1,
            ["2   ", "3   ", "4   ", "5   ", "6   "],
            None,
        )

        self.ltest(
            "trim none full focus",
            [T("1\n2\n3\n4\n5")],
            0,
            0,
            ["1   ", "2   ", "3   ", "4   ", "5   "],
            None,
        )

        self.ltest(
            "trim focus bottom",
            [T("1\n2\n3\n4\n5\n6")],
            0,
            0,
            ["1   ", "2   ", "3   ", "4   ", "5   "],
            None,
        )

        self.ltest(
            "trim focus top",
            [T("1\n2\n3\n4\n5\n6")],
            0,
            -1,
            ["2   ", "3   ", "4   ", "5   ", "6   "],
            None,
        )

        self.ltest(
            "trim focus top and bottom",
            [T("1\n2\n3\n4\n5\n6\n7")],
            0,
            -1,
            ["2   ", "3   ", "4   ", "5   ", "6   "],
            None,
        )

    def test3_shift(self):
        T, E = urwid.Text, urwid.Edit

        self.ltest(
            "shift up one fit",
            [T("1\n2"), T("3"), T("4"), T("5"), T("6")],
            4,
            5,
            ["2   ", "3   ", "4   ", "5   ", "6   "],
            None,
        )

        e = E("", "ab\nc", 1)
        e.set_edit_pos(2)
        self.ltest(
            "shift down one cursor over edge",
            [e, T("3"), T("4"), T("5\n6")],
            0,
            -1,
            ["ab  ", "c   ", "3   ", "4   ", "5   "],
            (2, 0),
        )

        self.ltest(
            "shift up one cursor over edge",
            [T("1\n2"), T("3"), T("4"), E("", "d\ne")],
            3,
            4,
            ["2   ", "3   ", "4   ", "d   ", "e   "],
            (1, 4),
        )

        self.ltest(
            "shift none cursor top focus over edge",
            [E("", "ab\n"), T("3"), T("4"), T("5\n6")],
            0,
            -1,
            ["    ", "3   ", "4   ", "5   ", "6   "],
            (0, 0),
        )

        e = E("", "abc\nd")
        e.set_edit_pos(3)
        self.ltest(
            "shift none cursor bottom focus over edge",
            [T("1\n2"), T("3"), T("4"), e],
            3,
            4,
            ["1   ", "2   ", "3   ", "4   ", "abc "],
            (3, 4),
        )

    def test4_really_large_contents(self):
        T, E = urwid.Text, urwid.Edit
        self.ltest(
            "really large edit",
            [T("hello" * 100)],
            0,
            0,
            ["hell", "ohel", "lohe", "lloh", "ello"],
            None,
        )

        self.ltest(
            "really large edit",
            [E("", "hello" * 100)],
            0,
            0,
            ["hell", "ohel", "lohe", "lloh", "llo "],
            (3, 4),
        )


class ListBoxKeypressTest(unittest.TestCase):
    def ktest(
        self,
        desc: str,
        key,
        body,
        focus: int,
        offset_inset: int,
        exp_focus: int,
        exp_offset_inset: int,
        exp_cur: tuple[int, int] | None,
        lbox=None,
    ) -> tuple[str | None, urwid.ListBox]:

        if lbox is None:
            lbox = urwid.ListBox(urwid.SimpleListWalker(body))
            lbox.body.set_focus(focus)
            lbox.shift_focus((4, 10), offset_inset)

        ret_key = lbox.keypress((4, 5), key)
        middle, top, bottom = lbox.calculate_visible((4, 5), focus=1)
        offset_inset, focus_widget, focus_pos, _ign, cursor = middle

        if cursor is not None:
            x, y = cursor
            y += offset_inset
            cursor = x, y

        exp = exp_focus, exp_offset_inset
        act = focus_pos, offset_inset
        self.assertEqual(exp, act)
        self.assertEqual(exp_cur, cursor)
        return ret_key, lbox

    def test1_up(self):
        T, S, E = urwid.Text, SelectableText, urwid.Edit

        self.ktest(
            "direct selectable both visible",
            "up",
            [S(""), S("")],
            1,
            1,
            0,
            0,
            None,
        )

        self.ktest(
            "selectable skip one all visible",
            "up",
            [S(""), T(""), S("")],
            2,
            2,
            0,
            0,
            None,
        )

        key, lbox = self.ktest(
            "nothing above no scroll",
            "up",
            [S("")],
            0,
            0,
            0,
            0,
            None,
        )
        self.assertEqual("up", key)

        key, lbox = self.ktest(
            "unselectable above no scroll",
            "up",
            [T(""), T(""), S("")],
            2,
            2,
            2,
            2,
            None,
        )
        self.assertEqual("up", key)

        self.ktest(
            "unselectable above scroll 1",
            "up",
            [T(""), S(""), T("\n\n\n")],
            1,
            0,
            1,
            1,
            None,
        )

        self.ktest(
            "selectable above scroll 1",
            "up",
            [S(""), S(""), T("\n\n\n")],
            1,
            0,
            0,
            0,
            None,
        )

        self.ktest(
            "selectable above too far",
            "up",
            [S(""), T(""), S(""), T("\n\n\n")],
            2,
            0,
            2,
            1,
            None,
        )

        self.ktest(
            "selectable above skip 1 scroll 1",
            "up",
            [S(""), T(""), S(""), T("\n\n\n")],
            2,
            1,
            0,
            0,
            None,
        )

        self.ktest(
            "tall selectable above scroll 2",
            "up",
            [S(""), S("\n"), S(""), T("\n\n\n")],
            2,
            0,
            1,
            0,
            None,
        )

        self.ktest(
            "very tall selectable above scroll 5",
            "up",
            [S(""), S("\n\n\n\n"), S(""), T("\n\n\n\n")],
            2,
            0,
            1,
            0,
            None,
        )

        self.ktest(
            "very tall selected scroll within 1",
            "up",
            [S(""), S("\n\n\n\n\n")],
            1,
            -1,
            1,
            0,
            None,
        )

        self.ktest(
            "edit above pass cursor",
            "up",
            [E("", "abc"), E("", "de")],
            1,
            1,
            0,
            0,
            (2, 0),
        )

        key, lbox = self.ktest(
            "edit too far above pass cursor A",
            "up",
            [E("", "abc"), T("\n\n\n\n"), E("", "de")],
            2,
            4,
            1,
            0,
            None,
        )

        self.ktest(
            "edit too far above pass cursor B",
            "up",
            None,
            None,
            None,
            0,
            0,
            (2, 0),
            lbox,
        )

        self.ktest(
            "within focus cursor made not visible",
            "up",
            [T("\n\n\n"), E("hi\n", "ab")],
            1,
            3,
            0,
            0,
            None,
        )

        self.ktest(
            "within focus cursor made not visible (2)",
            "up",
            [T("\n\n\n\n"), E("hi\n", "ab")],
            1,
            3,
            0,
            -1,
            None,
        )

        self.ktest(
            "force focus unselectable",
            "up",
            [T("\n\n\n\n"), S("")],
            1,
            4,
            0,
            0,
            None,
        )

        self.ktest(
            "pathological cursor widget",
            "up",
            [T("\n"), E("\n\n\n\n\n", "a")],
            1,
            4,
            0,
            -1,
            None,
        )

        self.ktest(
            "unselectable to unselectable",
            "up",
            [T(""), T(""), T(""), T(""), T(""), T(""), T("")],
            2,
            0,
            1,
            0,
            None,
        )

        self.ktest(
            "unselectable over edge to same",
            "up",
            [T(""), T("12\n34"), T(""), T(""), T(""), T("")],
            1,
            -1,
            1,
            0,
            None,
        )

        key, lbox = self.ktest(
            "edit short between pass cursor A",
            "up",
            [E("", "abcd"), E("", "a"), E("", "def")],
            2,
            2,
            1,
            1,
            (1, 1),
        )

        self.ktest(
            "edit short between pass cursor B",
            "up",
            None,
            None,
            None,
            0,
            0,
            (3, 0),
            lbox,
        )

        e = E("", "\n\n\n\n\n")
        e.set_edit_pos(1)
        key, lbox = self.ktest(
            "edit cursor force scroll",
            "up",
            [e],
            0,
            -1,
            0,
            0,
            (0, 0),
        )
        self.assertEqual(0, lbox.inset_fraction[0])

    def test2_down(self):
        T, S, E = urwid.Text, SelectableText, urwid.Edit

        self.ktest(
            "direct selectable both visible",
            "down",
            [S(""), S("")],
            0,
            0,
            1,
            1,
            None,
        )

        self.ktest(
            "selectable skip one all visible",
            "down",
            [S(""), T(""), S("")],
            0,
            0,
            2,
            2,
            None,
        )

        key, lbox = self.ktest(
            "nothing below no scroll",
            "down",
            [S("")],
            0,
            0,
            0,
            0,
            None,
        )
        self.assertEqual("down", key)

        key, lbox = self.ktest(
            "unselectable below no scroll",
            "down",
            [S(""), T(""), T("")],
            0,
            0,
            0,
            0,
            None,
        )
        self.assertEqual("down", key)

        self.ktest(
            "unselectable below scroll 1",
            "down",
            [T("\n\n\n"), S(""), T("")],
            1,
            4,
            1,
            3,
            None,
        )

        self.ktest(
            "selectable below scroll 1",
            "down",
            [T("\n\n\n"), S(""), S("")],
            1,
            4,
            2,
            4,
            None,
        )

        self.ktest(
            "selectable below too far",
            "down",
            [T("\n\n\n"), S(""), T(""), S("")],
            1,
            4,
            1,
            3,
            None,
        )

        self.ktest(
            "selectable below skip 1 scroll 1",
            "down",
            [T("\n\n\n"), S(""), T(""), S("")],
            1,
            3,
            3,
            4,
            None,
        )

        self.ktest(
            "tall selectable below scroll 2",
            "down",
            [T("\n\n\n"), S(""), S("\n"), S("")],
            1,
            4,
            2,
            3,
            None,
        )

        self.ktest(
            "very tall selectable below scroll 5",
            "down",
            [T("\n\n\n\n"), S(""), S("\n\n\n\n"), S("")],
            1,
            4,
            2,
            0,
            None,
        )

        self.ktest(
            "very tall selected scroll within 1",
            "down",
            [S("\n\n\n\n\n"), S("")],
            0,
            0,
            0,
            -1,
            None,
        )

        self.ktest(
            "edit below pass cursor",
            "down",
            [E("", "de"), E("", "abc")],
            0,
            0,
            1,
            1,
            (2, 1),
        )

        key, lbox = self.ktest(
            "edit too far below pass cursor A",
            "down",
            [E("", "de"), T("\n\n\n\n"), E("", "abc")],
            0,
            0,
            1,
            0,
            None,
        )

        self.ktest(
            "edit too far below pass cursor B",
            "down",
            None,
            None,
            None,
            2,
            4,
            (2, 4),
            lbox,
        )

        odd_e = E("", "hi\nab")
        odd_e.set_edit_pos(2)
        # disble cursor movement in odd_e object
        odd_e.move_cursor_to_coords = lambda s, c, xy: 0
        self.ktest(
            "within focus cursor made not visible",
            "down",
            [odd_e, T("\n\n\n\n")],
            0,
            0,
            1,
            1,
            None,
        )

        self.ktest(
            "within focus cursor made not visible (2)",
            "down",
            [
                odd_e,
                T("\n\n\n\n"),
            ],
            0,
            0,
            1,
            1,
            None,
        )

        self.ktest(
            "force focus unselectable",
            "down",
            [S(""), T("\n\n\n\n")],
            0,
            0,
            1,
            0,
            None,
        )

        odd_e.set_edit_text("hi\n\n\n\n\n")
        self.ktest(
            "pathological cursor widget",
            "down",
            [odd_e, T("\n")],
            0,
            0,
            1,
            4,
            None,
        )

        self.ktest(
            "unselectable to unselectable",
            "down",
            [T(""), T(""), T(""), T(""), T(""), T(""), T("")],
            4,
            4,
            5,
            4,
            None,
        )

        self.ktest(
            "unselectable over edge to same",
            "down",
            [T(""), T(""), T(""), T(""), T("12\n34"), T("")],
            4,
            4,
            4,
            3,
            None,
        )

        key, lbox = self.ktest(
            "edit short between pass cursor A",
            "down",
            [E("", "abc"), E("", "a"), E("", "defg")],
            0,
            0,
            1,
            1,
            (1, 1),
        )

        self.ktest(
            "edit short between pass cursor B",
            "down",
            None,
            None,
            None,
            2,
            2,
            (3, 2),
            lbox,
        )

        e = E("", "\n\n\n\n\n")
        e.set_edit_pos(4)
        key, lbox = self.ktest(
            "edit cursor force scroll",
            "down",
            [e],
            0,
            0,
            0,
            -1,
            (0, 4),
        )
        self.assertEqual(1, lbox.inset_fraction[0])

    def test3_page_up(self):
        T, S, E = urwid.Text, SelectableText, urwid.Edit

        self.ktest(
            "unselectable aligned to aligned",
            "page up",
            [T(""), T("\n"), T("\n\n"), T(""), T("\n"), T("\n\n")],
            3,
            0,
            1,
            0,
            None,
        )

        self.ktest(
            "unselectable unaligned to aligned",
            "page up",
            [T(""), T("\n"), T("\n"), T("\n"), T("\n"), T("\n\n")],
            3,
            -1,
            1,
            0,
            None,
        )

        self.ktest(
            "selectable to unselectable",
            "page up",
            [T(""), T("\n"), T("\n"), T("\n"), S("\n"), T("\n\n")],
            4,
            1,
            1,
            -1,
            None,
        )

        self.ktest(
            "selectable to cut off selectable",
            "page up",
            [S("\n\n"), T("\n"), T("\n"), S("\n"), T("\n\n")],
            3,
            1,
            0,
            -1,
            None,
        )

        self.ktest(
            "seletable to selectable",
            "page up",
            [T("\n\n"), S("\n"), T("\n"), S("\n"), T("\n\n")],
            3,
            1,
            1,
            1,
            None,
        )

        self.ktest(
            "within very long selectable",
            "page up",
            [S(""), S("\n\n\n\n\n\n\n\n"), T("\n")],
            1,
            -6,
            1,
            -1,
            None,
        )

        e = E("", "\n\nab\n\n\n\n\ncd\n")
        e.set_edit_pos(11)
        self.ktest(
            "within very long cursor widget",
            "page up",
            [S(""), e, T("\n")],
            1,
            -6,
            1,
            -2,
            (2, 0),
        )

        self.ktest(
            "pathological cursor widget",
            "page up",
            [T(""), E("\n\n\n\n\n\n\n\n", "ab"), T("")],
            1,
            -5,
            0,
            0,
            None,
        )

        e = E("", "\nab\n\n\n\n\ncd\n")
        e.set_edit_pos(10)
        self.ktest(
            "very long cursor widget snap",
            "page up",
            [T(""), e, T("\n")],
            1,
            -5,
            1,
            0,
            (2, 1),
        )

        self.ktest(
            "slight scroll selectable",
            "page up",
            [T("\n"), S("\n"), T(""), S(""), T("\n\n\n"), S("")],
            5,
            4,
            3,
            0,
            None,
        )

        self.ktest(
            "scroll into snap region",
            "page up",
            [T("\n"), S("\n"), T(""), T(""), T("\n\n\n"), S("")],
            5,
            4,
            1,
            0,
            None,
        )

        self.ktest(
            "mid scroll short",
            "page up",
            [T("\n"), T(""), T(""), S(""), T(""), T("\n"), S(""), T("\n")],
            6,
            2,
            3,
            1,
            None,
        )

        self.ktest(
            "mid scroll long",
            "page up",
            [T("\n"), S(""), T(""), S(""), T(""), T("\n"), S(""), T("\n")],
            6,
            2,
            1,
            0,
            None,
        )

        self.ktest(
            "mid scroll perfect",
            "page up",
            [T("\n"), S(""), S(""), S(""), T(""), T("\n"), S(""), T("\n")],
            6,
            2,
            2,
            0,
            None,
        )

        self.ktest(
            "cursor move up fail short",
            "page up",
            [T("\n"), T("\n"), E("", "\nab"), T(""), T("")],
            2,
            1,
            2,
            4,
            (0, 4),
        )

        self.ktest(
            "cursor force fail short",
            "page up",
            [T("\n"), T("\n"), E("\n", "ab"), T(""), T("")],
            2,
            1,
            0,
            0,
            None,
        )

        odd_e = E("", "hi\nab")
        odd_e.set_edit_pos(2)
        # disble cursor movement in odd_e object
        odd_e.move_cursor_to_coords = lambda s, c, xy: 0
        self.ktest(
            "cursor force fail long",
            "page up",
            [odd_e, T("\n"), T("\n"), T("\n"), S(""), T("\n")],
            4,
            2,
            1,
            -1,
            None,
        )

        self.ktest(
            "prefer not cut off",
            "page up",
            [S("\n"), T("\n"), S(""), T("\n\n"), S(""), T("\n")],
            4,
            2,
            2,
            1,
            None,
        )

        self.ktest(
            "allow cut off",
            "page up",
            [S("\n"), T("\n"), T(""), T("\n\n"), S(""), T("\n")],
            4,
            2,
            0,
            -1,
            None,
        )

        self.ktest(
            "at top fail",
            "page up",
            [T("\n\n"), T("\n"), T("\n\n\n")],
            0,
            0,
            0,
            0,
            None,
        )

        self.ktest(
            "all visible fail",
            "page up",
            [T("a"), T("\n")],
            0,
            0,
            0,
            0,
            None,
        )

        self.ktest(
            "current ok fail",
            "page up",
            [T("\n\n"), S("hi")],
            1,
            3,
            1,
            3,
            None,
        )

        self.ktest(
            "all visible choose top selectable",
            "page up",
            [T(""), S("a"), S("b"), S("c")],
            3,
            3,
            1,
            1,
            None,
        )

        self.ktest(
            "bring in edge choose top",
            "page up",
            [S("b"), T("-"), S("-"), T("c"), S("d"), T("-")],
            4,
            3,
            0,
            0,
            None,
        )

        self.ktest(
            "bring in edge choose top selectable",
            "page up",
            [T("b"), S("-"), S("-"), T("c"), S("d"), T("-")],
            4,
            3,
            1,
            1,
            None,
        )

    def test4_page_down(self):
        T, S, E = urwid.Text, SelectableText, urwid.Edit

        self.ktest(
            "unselectable aligned to aligned",
            "page down",
            [T("\n\n"), T("\n"), T(""), T("\n\n"), T("\n"), T("")],
            2,
            4,
            4,
            3,
            None,
        )

        self.ktest(
            "unselectable unaligned to aligned",
            "page down",
            [T("\n\n"), T("\n"), T("\n"), T("\n"), T("\n"), T("")],
            2,
            4,
            4,
            3,
            None,
        )

        self.ktest(
            "selectable to unselectable",
            "page down",
            [T("\n\n"), S("\n"), T("\n"), T("\n"), T("\n"), T("")],
            1,
            2,
            4,
            4,
            None,
        )

        self.ktest(
            "selectable to cut off selectable",
            "page down",
            [T("\n\n"), S("\n"), T("\n"), T("\n"), S("\n\n")],
            1,
            2,
            4,
            3,
            None,
        )

        self.ktest(
            "seletable to selectable",
            "page down",
            [T("\n\n"), S("\n"), T("\n"), S("\n"), T("\n\n")],
            1,
            1,
            3,
            2,
            None,
        )

        self.ktest(
            "within very long selectable",
            "page down",
            [T("\n"), S("\n\n\n\n\n\n\n\n"), S("")],
            1,
            2,
            1,
            -3,
            None,
        )

        e = E("", "\nab\n\n\n\n\ncd\n\n")
        e.set_edit_pos(2)
        self.ktest(
            "within very long cursor widget",
            "page down",
            [T("\n"), e, S("")],
            1,
            2,
            1,
            -2,
            (1, 4),
        )

        odd_e = E("", "ab\n\n\n\n\n\n\n\n\n")
        odd_e.set_edit_pos(1)
        # disble cursor movement in odd_e object
        odd_e.move_cursor_to_coords = lambda s, c, xy: 0
        self.ktest(
            "pathological cursor widget",
            "page down",
            [T(""), odd_e, T("")],
            1,
            1,
            2,
            4,
            None,
        )

        e = E("", "\nab\n\n\n\n\ncd\n")
        e.set_edit_pos(2)
        self.ktest(
            "very long cursor widget snap",
            "page down",
            [T("\n"), e, T("")],
            1,
            2,
            1,
            -3,
            (1, 3),
        )

        self.ktest(
            "slight scroll selectable",
            "page down",
            [S(""), T("\n\n\n"), S(""), T(""), S("\n"), T("\n")],
            0,
            0,
            2,
            4,
            None,
        )

        self.ktest(
            "scroll into snap region",
            "page down",
            [S(""), T("\n\n\n"), T(""), T(""), S("\n"), T("\n")],
            0,
            0,
            4,
            3,
            None,
        )

        self.ktest(
            "mid scroll short",
            "page down",
            [T("\n"), S(""), T("\n"), T(""), S(""), T(""), T(""), T("\n")],
            1,
            2,
            4,
            3,
            None,
        )

        self.ktest(
            "mid scroll long",
            "page down",
            [T("\n"), S(""), T("\n"), T(""), S(""), T(""), S(""), T("\n")],
            1,
            2,
            6,
            4,
            None,
        )

        self.ktest(
            "mid scroll perfect",
            "page down",
            [T("\n"), S(""), T("\n"), T(""), S(""), S(""), S(""), T("\n")],
            1,
            2,
            5,
            4,
            None,
        )

        e = E("", "hi\nab")
        e.set_edit_pos(1)
        self.ktest(
            "cursor move up fail short",
            "page down",
            [T(""), T(""), e, T("\n"), T("\n")],
            2,
            1,
            2,
            -1,
            (1, 0),
        )

        odd_e = E("", "hi\nab")
        odd_e.set_edit_pos(1)
        # disble cursor movement in odd_e object
        odd_e.move_cursor_to_coords = lambda s, c, xy: 0
        self.ktest(
            "cursor force fail short",
            "page down",
            [T(""), T(""), odd_e, T("\n"), T("\n")],
            2,
            2,
            4,
            3,
            None,
        )

        self.ktest(
            "cursor force fail long",
            "page down",
            [T("\n"), S(""), T("\n"), T("\n"), T("\n"), E("hi\n", "ab")],
            1,
            2,
            4,
            4,
            None,
        )

        self.ktest(
            "prefer not cut off",
            "page down",
            [T("\n"), S(""), T("\n\n"), S(""), T("\n"), S("\n")],
            1,
            2,
            3,
            3,
            None,
        )

        self.ktest(
            "allow cut off",
            "page down",
            [T("\n"), S(""), T("\n\n"), T(""), T("\n"), S("\n")],
            1,
            2,
            5,
            4,
            None,
        )

        self.ktest(
            "at bottom fail",
            "page down",
            [T("\n\n"), T("\n"), T("\n\n\n")],
            2,
            1,
            2,
            1,
            None,
        )

        self.ktest(
            "all visible fail",
            "page down",
            [T("a"), T("\n")],
            1,
            1,
            1,
            1,
            None,
        )

        self.ktest(
            "current ok fail",
            "page down",
            [S("hi"), T("\n\n")],
            0,
            0,
            0,
            0,
            None,
        )

        self.ktest(
            "all visible choose last selectable",
            "page down",
            [S("a"), S("b"), S("c"), T("")],
            0,
            0,
            2,
            2,
            None,
        )

        self.ktest(
            "bring in edge choose last",
            "page down",
            [T("-"), S("d"), T("c"), S("-"), T("-"), S("b")],
            1,
            1,
            5,
            4,
            None,
        )

        self.ktest(
            "bring in edge choose last selectable",
            "page down",
            [T("-"), S("d"), T("c"), S("-"), S("-"), T("b")],
            1,
            1,
            4,
            3,
            None,
        )


class ZeroHeightContentsTest(unittest.TestCase):
    def test_listbox_pile(self):
        lb = urwid.ListBox(urwid.SimpleListWalker([urwid.Pile([])]))
        size = (2, 5)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"  ", b"  ", b"  ", b"  ", b"  "], canvas.text)

    def test_listbox_text_pile_page_down(self):
        lb = urwid.ListBox(urwid.SimpleListWalker([urwid.Text("above"), urwid.Pile([])]))
        size = (7, 5)
        lb.keypress(size, "page down")
        self.assertEqual(lb.focus_position, 0)
        lb.keypress(size, "page down")  # second one caused ListBox failure
        self.assertEqual(lb.focus_position, 0)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"above  ", b"       ", b"       ", b"       ", b"       "], canvas.text)

    def test_listbox_text_pile_page_up(self):
        lb = urwid.ListBox(urwid.SimpleListWalker([urwid.Pile([]), urwid.Text("below")]))
        size = (7, 5)
        lb.set_focus(1)
        lb.keypress(size, "page up")
        self.assertEqual(lb.focus_position, 1)
        lb.keypress(size, "page up")  # second one caused pile failure
        self.assertEqual(lb.focus_position, 1)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"below  ", b"       ", b"       ", b"       ", b"       "], canvas.text)

    def test_listbox_text_pile_down(self):
        sp = urwid.Pile([])
        sp.selectable = lambda: True  # abuse our Pile
        lb = urwid.ListBox(urwid.SimpleListWalker([urwid.Text("above"), sp]))
        size = (7, 5)
        lb.keypress(size, "down")
        self.assertEqual(lb.focus_position, 0)
        lb.keypress(size, "down")
        self.assertEqual(lb.focus_position, 0)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"above  ", b"       ", b"       ", b"       ", b"       "], canvas.text)

    def test_listbox_text_pile_up(self):
        sp = urwid.Pile([])
        sp.selectable = lambda: True  # abuse our Pile
        lb = urwid.ListBox(urwid.SimpleListWalker([sp, urwid.Text("below")]))
        size = (7, 5)
        lb.set_focus(1)
        lb.keypress(size, "up")
        self.assertEqual(lb.focus_position, 1)
        lb.keypress(size, "up")
        self.assertEqual(lb.focus_position, 1)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"below  ", b"       ", b"       ", b"       ", b"       "], canvas.text)


class PageDownAboveTopTest(unittest.TestCase):
    def test_page_down_does_not_raise(self):
        # Two-row rows so the four-item list does not fit the viewport at once.
        lb = urwid.ListBox(
            urwid.SimpleListWalker(
                [
                    urwid.Text("a0\nb0"),
                    SelectableText("a1\nb1"),
                    urwid.Text("a2\nb2"),
                    urwid.Text("a3\nb3"),
                ]
            )
        )
        size = (7, 4)
        lb.set_focus(0)
        lb.render(size, focus=True)
        lb.keypress(size, "page down")  # used to raise ListBoxError
        self.assertEqual(lb.focus_position, 3)
        canvas = lb.render(size, focus=True)
        self.assertEqual([b"a2     ", b"b2     ", b"a3     ", b"b3     "], canvas.text)
        # a second page down at the bottom is a no-op, not a crash
        lb.keypress(size, "page down")
        self.assertEqual(lb.focus_position, 3)


class ListBoxSetBodyTest(unittest.TestCase):
    def test_signal_connected(self):
        lb = urwid.ListBox([])
        lb.body = urwid.SimpleListWalker([])
        self.assertEqual(
            lb.body._urwid_signals["modified"][0][1],
            lb._invalidate,
            "outdated canvas cache reuse after ListWalker's contents modified",
        )


class TestListWalkerFromIterable(unittest.TestCase):
    def test_01_simple_list_walker(self):
        walker = urwid.SimpleListWalker(str(num) for num in range(5))
        self.assertEqual(5, len(walker))

    def test_02_simple_focus_list_walker(self):
        walker = urwid.SimpleFocusListWalker(str(num) for num in range(5))
        self.assertEqual(5, len(walker))



from __future__ import annotations

import string
import typing
import unittest

import urwid

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

LGPL_HEADER = """
Copyright (C) <year>  <name of author>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
"""


class TestScrollable(unittest.TestCase):
    def test_basic(self):
        """Test basic init and scroll."""
        long_content = urwid.Text(LGPL_HEADER)
        reduced_size = (80, 5)
        content_size = long_content.pack()

        widget = urwid.Scrollable(long_content)
        self.assertEqual(frozenset((urwid.BOX,)), widget.sizing())

        long_content_canvas = long_content.render((reduced_size[0],))
        cropped_content_canvas = urwid.CompositeCanvas(long_content_canvas)
        cropped_content_canvas.trim_end(content_size[1] - reduced_size[1])
        top_decoded = cropped_content_canvas.decoded_text

        self.assertEqual(
            top_decoded,
            widget.render(reduced_size).decoded_text,
        )

        for key_1, key_2 in (("down", "up"), ("page down", "page up"), ("end", "home")):
            widget.keypress(reduced_size, key_1)

            self.assertNotEqual(top_decoded, widget.render(reduced_size).decoded_text)

            widget.keypress(reduced_size, key_2)

            self.assertEqual(top_decoded, widget.render(reduced_size).decoded_text)

    def test_negative(self):
        with self.assertRaises(ValueError):
            urwid.Scrollable(urwid.SolidFill(" "))


class TestScrollBarScrollable(unittest.TestCase):
    def test_basic(self):
        """Test basic init and scroll.

        Unlike `Scrollable`, `ScrollBar` can be also scrolled with a mouse wheel.
        """

        long_content = urwid.Text(LGPL_HEADER)
        reduced_size = (40, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(long_content))

        self.assertEqual(frozenset((urwid.BOX,)), widget.sizing())

        top_position_rendered = (
            "                                       █",
            "Copyright (C) <year>  <name of author>  ",
            "                                        ",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
        )
        pos_1_down_rendered = (
            "Copyright (C) <year>  <name of author>  ",
            "                                       █",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
            "modify it under the terms of the GNU    ",
        )

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            (
                "redistribute it and/or                  ",
                "modify it under the terms of the GNU   █",
                "Lesser General Public                   ",
                "License as published by the Free        ",
                "Software Foundation; either             ",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "end")

        self.assertEqual(
            (
                "not, write to the Free Software         ",
                "Foundation, Inc., 51 Franklin Street,   ",
                "Fifth Floor, Boston, MA  02110-1301     ",
                "USA                                     ",
                "                                       █",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "home")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 5, 1, 1, False))

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 4, 1, 1, False))

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

    def test_mouse_left_click_scrollbar(self):
        """Left click on the rendered scrollbar jumps to the clicked position."""
        content = urwid.Text("\n".join(string.ascii_letters))  # 52 single-char lines
        reduced_size = (3, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(content))
        scrollable = widget.original_widget
        scrollbar_col = reduced_size[0] - 1  # right side scrollbar

        self.assertEqual(0, scrollable.get_scrollpos())

        # posmax == 52 - 5 == 47, thumb_height == 1, thumb travel == 5 - 1 == 4.
        # Clicking the bottom row moves past the end and is clamped to posmax.
        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 1, scrollbar_col, 4, False))
        self.assertEqual(47, scrollable.get_scrollpos())

        # Clicking the top row scrolls back to the top.
        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 1, scrollbar_col, 0, False))
        self.assertEqual(0, scrollable.get_scrollpos())

        # Intermediate rows map proportionally: round(row * posmax / travel).
        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 1, scrollbar_col, 1, False))
        self.assertEqual(12, scrollable.get_scrollpos())

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 1, scrollbar_col, 3, False))
        self.assertEqual(35, scrollable.get_scrollpos())

    def test_mouse_left_click_content_ignored(self):
        """Left click outside of the scrollbar columns does not scroll."""
        content = urwid.Text("\n".join(string.ascii_letters))
        reduced_size = (3, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(content))
        scrollable = widget.original_widget

        self.assertEqual(0, scrollable.get_scrollpos())

        # Column 0 belongs to the wrapped content, not the scrollbar.
        self.assertFalse(widget.mouse_event(reduced_size, "mouse press", 1, 0, 4, False))
        self.assertEqual(0, scrollable.get_scrollpos())

    def test_mouse_left_click_left_side_scrollbar(self):
        """Left click detection honours a left-aligned scrollbar."""
        content = urwid.Text("\n".join(string.ascii_letters))
        reduced_size = (3, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(content), side="left")
        scrollable = widget.original_widget

        self.assertEqual(0, scrollable.get_scrollpos())

        # The scrollbar occupies column 0 when aligned to the left.
        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 1, 0, 4, False))
        self.assertEqual(47, scrollable.get_scrollpos())

        # A click on the content columns is ignored.
        self.assertFalse(widget.mouse_event(reduced_size, "mouse press", 1, 1, 0, False))
        self.assertEqual(47, scrollable.get_scrollpos())

    def test_mouse_left_click_without_scrollbar(self):
        """Without a rendered scrollbar a left click is not handled."""
        content = urwid.Text("a\nb")  # fits into the available height
        reduced_size = (3, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(content))
        scrollable = widget.original_widget

        self.assertFalse(widget.mouse_event(reduced_size, "mouse press", 1, 2, 2, False))
        self.assertEqual(0, scrollable.get_scrollpos())

    def test_alt_symbols(self):
        long_content = urwid.Text(LGPL_HEADER)
        reduced_size = (40, 5)
        widget = urwid.ScrollBar(
            urwid.Scrollable(long_content),
            trough_char=urwid.ScrollBar.Symbols.LITE_SHADE,
        )

        self.assertEqual(
            (
                "                                       █",
                "Copyright (C) <year>  <name of author> ░",
                "                                       ░",
                "This library is free software; you can ░",
                "redistribute it and/or                 ░",
            ),
            widget.render(reduced_size).decoded_text,
        )

    def test_fixed(self):
        """Test with fixed wrapped widget."""
        widget = urwid.ScrollBar(
            urwid.Scrollable(urwid.BigText("1", urwid.HalfBlockHeavy6x5Font())),
            trough_char=urwid.ScrollBar.Symbols.LITE_SHADE,
            thumb_char=urwid.ScrollBar.Symbols.DARK_SHADE,
        )
        reduced_size = (8, 3)

        self.assertEqual(
            (
                " ▐█▌   ▓",
                " ▀█▌   ▓",
                "  █▌   ░",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            (
                "  █▌   ░",
                "  █▌   ▓",
                " ███▌  ▓",
            ),
            widget.render(reduced_size).decoded_text,
        )

    def test_negative(self):
        with self.assertRaises(ValueError):
            urwid.ScrollBar(urwid.Text(" "))

        with self.assertRaises(TypeError):
            urwid.ScrollBar(urwid.SolidFill(" "))

    def test_no_scrollbar(self):
        """If widget fit without scroll - no scrollbar needed"""
        widget = urwid.ScrollBar(
            urwid.Scrollable(urwid.BigText("1", urwid.HalfBlockHeavy6x5Font())),
            trough_char=urwid.ScrollBar.Symbols.LITE_SHADE,
            thumb_char=urwid.ScrollBar.Symbols.DARK_SHADE,
        )
        reduced_size = (8, 5)
        self.assertEqual(
            (
                " ▐█▌    ",
                " ▀█▌    ",
                "  █▌    ",
                "  █▌    ",
                " ███▌   ",
            ),
            widget.render(reduced_size).decoded_text,
        )


class TestScrollBarListBox(unittest.TestCase):
    def test_relative_non_selectable(self):
        widget = urwid.ScrollBar(
            urwid.ListBox(urwid.SimpleListWalker(urwid.Text(line) for line in LGPL_HEADER.splitlines()))
        )

        reduced_size = (40, 5)

        top_position_rendered = (
            "                                       █",
            "Copyright (C) <year>  <name of author>  ",
            "                                        ",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
        )
        pos_1_down_rendered = (
            "Copyright (C) <year>  <name of author>  ",
            "                                       █",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
            "modify it under the terms of the GNU    ",
        )

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            (
                "modify it under the terms of the GNU    ",
                "Lesser General Public                  █",
                "License as published by the Free        ",
                "Software Foundation; either             ",
                "version 2.1 of the License, or (at your ",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "end")

        self.assertEqual(
            (
                "License along with this library; if     ",
                "not, write to the Free Software         ",
                "Foundation, Inc., 51 Franklin Street,   ",
                "Fifth Floor, Boston, MA  02110-1301     ",
                "USA                                    █",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "home")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 5, 1, 1, False))

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 4, 1, 1, False))

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

    def test_empty(self):
        """Empty widget should be correctly rendered."""
        widget = urwid.ScrollBar(urwid.ListBox(urwid.SimpleListWalker(())))
        reduced_size = (10, 5)
        self.assertEqual(
            (
                "          ",
                "          ",
                "          ",
                "          ",
                "          ",
            ),
            widget.render(reduced_size).decoded_text,
        )

    def test_minimal_height(self):
        """If we have only 1 line render height and thumb position in the middle - do not render top."""
        widget = urwid.ScrollBar(
            urwid.ListBox(
                (
                    urwid.CheckBox("A"),
                    urwid.CheckBox("B"),
                    urwid.CheckBox("C"),
                )
            )
        )
        reduced_size = (7, 1)
        self.assertEqual(("[ ] A █",), widget.render(reduced_size).decoded_text)
        widget.keypress(reduced_size, "down")
        self.assertEqual(("[ ] B █",), widget.render(reduced_size).decoded_text)
        widget.keypress(reduced_size, "down")
        self.assertEqual(("[ ] C █",), widget.render(reduced_size).decoded_text)


def trivial_AttrMap(widget):
    return urwid.AttrMap(widget, {})


class TestScrollableAttrMap(unittest.TestCase):
    def test_basic(self):
        """Test basic init and scroll."""
        long_content = urwid.Text(LGPL_HEADER)
        reduced_size = (80, 5)
        content_size = long_content.pack()

        widget = urwid.Scrollable(trivial_AttrMap(long_content))
        self.assertEqual(frozenset((urwid.BOX,)), widget.sizing())

        cropped_content_canvas = urwid.CompositeCanvas(long_content.render((reduced_size[0],)))
        cropped_content_canvas.trim_end(content_size[1] - reduced_size[1])
        top_decoded = cropped_content_canvas.decoded_text

        self.assertEqual(
            top_decoded,
            widget.render(reduced_size).decoded_text,
        )

        for key_1, key_2 in (("down", "up"), ("page down", "page up"), ("end", "home")):
            widget.keypress(reduced_size, key_1)

            self.assertNotEqual(top_decoded, widget.render(reduced_size).decoded_text)

            widget.keypress(reduced_size, key_2)

            self.assertEqual(top_decoded, widget.render(reduced_size).decoded_text)

    def test_negative(self):
        with self.assertRaises(ValueError):
            urwid.Scrollable(trivial_AttrMap(urwid.SolidFill(" ")))


class TestScrollBarAttrMap(unittest.TestCase):
    def test_basic(self):
        """Test basic init and scroll.

        Unlike `Scrollable`, `ScrollBar` can be also scrolled with a mouse wheel.
        """

        long_content = urwid.Text(LGPL_HEADER)
        reduced_size = (40, 5)
        widget = urwid.ScrollBar(urwid.Scrollable(trivial_AttrMap(long_content)))

        self.assertEqual(frozenset((urwid.BOX,)), widget.sizing())

        top_position_rendered = (
            "                                       █",
            "Copyright (C) <year>  <name of author>  ",
            "                                        ",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
        )
        pos_1_down_rendered = (
            "Copyright (C) <year>  <name of author>  ",
            "                                       █",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
            "modify it under the terms of the GNU    ",
        )

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            (
                "redistribute it and/or                  ",
                "modify it under the terms of the GNU   █",
                "Lesser General Public                   ",
                "License as published by the Free        ",
                "Software Foundation; either             ",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "end")

        self.assertEqual(
            (
                "not, write to the Free Software         ",
                "Foundation, Inc., 51 Franklin Street,   ",
                "Fifth Floor, Boston, MA  02110-1301     ",
                "USA                                     ",
                "                                       █",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "home")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 5, 1, 1, False))

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 4, 1, 1, False))

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)


class TestScrollBarListBoxAttrMap(unittest.TestCase):
    def test_relative_non_selectable(self):
        widget = urwid.ScrollBar(
            trivial_AttrMap(
                urwid.ListBox(urwid.SimpleListWalker(urwid.Text(line) for line in LGPL_HEADER.splitlines()))
            )
        )

        reduced_size = (40, 5)

        top_position_rendered = (
            "                                       █",
            "Copyright (C) <year>  <name of author>  ",
            "                                        ",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
        )
        pos_1_down_rendered = (
            "Copyright (C) <year>  <name of author>  ",
            "                                       █",
            "This library is free software; you can  ",
            "redistribute it and/or                  ",
            "modify it under the terms of the GNU    ",
        )

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            (
                "modify it under the terms of the GNU    ",
                "Lesser General Public                  █",
                "License as published by the Free        ",
                "Software Foundation; either             ",
                "version 2.1 of the License, or (at your ",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "end")

        self.assertEqual(
            (
                "License along with this library; if     ",
                "not, write to the Free Software         ",
                "Foundation, Inc., 51 Franklin Street,   ",
                "Fifth Floor, Boston, MA  02110-1301     ",
                "USA                                    █",
            ),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "home")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 5, 1, 1, False))

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 4, 1, 1, False))

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

    def test_large_non_selectable(self):
        top = urwid.Text("\n".join(string.ascii_letters))
        bottom = urwid.Text("\n".join(string.digits))
        widget = urwid.ScrollBar(urwid.ListBox(urwid.SimpleListWalker((top, bottom))))

        reduced_size = (3, 5)

        top_position_rendered = ("a █", "b  ", "c  ", "d  ", "e  ")
        pos_1_down_rendered = ("b  ", "c █", "d  ", "e  ", "f  ")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "page down")

        self.assertEqual(
            ("f  ", "g █", "h  ", "i  ", "j  "),
            widget.render(reduced_size).decoded_text,
        )

        widget.keypress(reduced_size, "page up")

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 5, 1, 1, False))

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

        self.assertTrue(widget.mouse_event(reduced_size, "mouse press", 4, 1, 1, False))

        self.assertEqual(top_position_rendered, widget.render(reduced_size).decoded_text)

    def test_large_selectable(self):
        """Input is handled by LineBox and wrapped widgets."""
        top = urwid.Edit("\n".join(string.ascii_letters))
        bottom = urwid.IntEdit("\n".join(string.digits))
        widget = urwid.ScrollBar(urwid.ListBox(urwid.SimpleListWalker((top, bottom))))

        reduced_size = (3, 10)

        top_position_rendered = ("a █", "b █", "c  ", "d  ", "e  ", "f  ", "g  ", "h  ", "i  ", "j  ")
        pos_1_down_rendered = ("0  ", "1  ", "2  ", "3  ", "4  ", "5  ", "6  ", "7  ", "8 █", "9 █")

        self.assertEqual((top_position_rendered), widget.render(reduced_size).decoded_text)

        widget.keypress(reduced_size, "down")

        self.assertEqual(pos_1_down_rendered, widget.render(reduced_size).decoded_text)

    def test_hinted_len(self):
        class HintedWalker(urwid.ListWalker):
            def __init__(self, items: Iterable[str]) -> None:
                self.items: tuple[str] = tuple(items)
                self.focus = 0
                self.requested_numbers: set[int] = set()

            def __length_hint__(self) -> int:
                return len(self.items)

            def __getitem__(self, item: int) -> urwid.Text:
                self.requested_numbers.add(item)
                return urwid.Text(self.items[item])

            def set_focus(self, item: int) -> None:
                self.focus = item

            def next_position(self, position: int) -> int:
                if position + 1 < len(self.items):
                    return position + 1
                raise IndexError

            def prev_position(self, position: int) -> int:
                if position - 1 >= 0:
                    return position - 1
                raise IndexError

        widget = urwid.ScrollBar(urwid.ListBox(HintedWalker((f"Line {idx:02}") for idx in range(1, 51))))
        size = (10, 10)
        widget.original_widget.focus_position = 19
        self.assertEqual(
            (
                "Line 16   ",
                "Line 17   ",
                "Line 18   ",
                "Line 19  █",
                "Line 20  █",
                "Line 21   ",
                "Line 22   ",
                "Line 23   ",
                "Line 24   ",
                "Line 25   ",
            ),
            widget.render(size).decoded_text,
        )
        self.assertNotIn(
            30,
            widget.original_widget.body.requested_numbers,
            "Requested index out of range [0, last shown]. This means not relative scroll bar built.",
        )



from __future__ import annotations

import typing

import urwid

palette = [("I say", "default,bold", "default", "bold")]
ask = urwid.Edit(("I say", "What is your name?\n"))
reply = urwid.Text("")
button_inst = urwid.Button("Exit")
div = urwid.Divider()
pile = urwid.Pile([ask, div, reply, div, button_inst])
top = urwid.Filler(pile, valign="top")


def on_ask_change(_edit: urwid.Edit, new_edit_text: str) -> None:
    reply.set_text(("I say", f"Nice to meet you, {new_edit_text}"))


def on_exit_clicked(_button: urwid.Button) -> typing.NoReturn:
    raise urwid.ExitMainLoop()


urwid.connect_signal(ask, "change", on_ask_change)
urwid.connect_signal(button_inst, "click", on_exit_clicked)

urwid.MainLoop(top, palette).run()



from __future__ import annotations

import urwid


class QuestionnaireItem(urwid.WidgetWrap[urwid.GridFlow]):
    def __init__(self) -> None:
        self.options: list[urwid.RadioButton] = []
        unsure = urwid.RadioButton(self.options, "Unsure")
        yes = urwid.RadioButton(self.options, "Yes")
        no = urwid.RadioButton(self.options, "No")
        display_widget = urwid.GridFlow([unsure, yes, no], 15, 3, 1, "left")
        super().__init__(display_widget)

    def get_state(self) -> str:
        return next(o.label for o in self.options if o.state is True)



from __future__ import annotations

import asyncio
import concurrent.futures
import socket
import sys
import typing
import unittest

import urwid

if typing.TYPE_CHECKING:
    from concurrent.futures import Future
    from types import TracebackType

try:
    import gi.repository
except ImportError:
    GLIB_AVAILABLE = False
else:
    GLIB_AVAILABLE = True

try:
    import tornado
except ImportError:
    TORNADO_AVAILABLE = False
else:
    TORNADO_AVAILABLE = True

try:
    import twisted
except ImportError:
    TWISTED_AVAILABLE = False
else:
    TWISTED_AVAILABLE = True

try:
    import trio
except ImportError:
    TRIO_AVAILABLE = False
else:
    TRIO_AVAILABLE = True

try:
    import zmq
except ImportError:
    ZMQ_AVAILABLE = False
else:
    ZMQ_AVAILABLE = True

IS_WINDOWS = sys.platform == "win32"


class ClosingSocketPair(typing.ContextManager[tuple[socket.socket, socket.socket]]):
    __slots__ = ("rd_s", "wr_s")

    def __init__(self) -> None:
        self.rd_s: socket.socket | None = None
        self.wr_s: socket.socket | None = None

    def __enter__(self) -> tuple[socket.socket, socket.socket]:
        self.rd_s, self.wr_s = socket.socketpair()
        return self.rd_s, self.wr_s

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self.rd_s is not None:
            self.rd_s.close()
        if self.wr_s is not None:
            self.wr_s.close()


class EventLoopTestMixin:
    evl: urwid.EventLoop

    def test_event_loop(self):
        evl: urwid.EventLoop = self.evl
        out = []
        rd: socket.socket
        wr: socket.socket

        def step1():
            out.append("writing")
            wr.send(b"hi")

        def step2():
            out.append(rd.recv(2).decode("ascii"))
            raise urwid.ExitMainLoop

        with ClosingSocketPair() as (rd, wr):
            _handle = evl.alarm(0, step1)
            _handle = evl.watch_file(rd.fileno(), step2)

            evl.run()

        self.assertEqual(out, ["writing", "hi"])

    def test_remove_alarm(self):
        evl: urwid.EventLoop = self.evl
        handle = evl.alarm(50, lambda: None)

        def step1():
            self.assertTrue(evl.remove_alarm(handle))
            self.assertFalse(evl.remove_alarm(handle))
            raise urwid.ExitMainLoop

        evl.alarm(0, step1)
        evl.run()

    def test_remove_watch_file(self):
        evl: urwid.EventLoop = self.evl

        with ClosingSocketPair() as (rd, _wr):
            handle = evl.watch_file(rd.fileno(), lambda: None)

            def step1():
                self.assertTrue(evl.remove_watch_file(handle))
                self.assertFalse(evl.remove_watch_file(handle))
                raise urwid.ExitMainLoop

            evl.alarm(0, step1)
            evl.run()

    _expected_idle_handle = 1

    def test_run(self):
        evl: urwid.EventLoop = self.evl
        out = []
        wr: socket.socket
        rd: socket.socket

        def say_hello():
            out.append("hello")

        def say_waiting():
            out.append("waiting")

        def exit_clean() -> typing.NoReturn:
            out.append("clean exit")
            raise urwid.ExitMainLoop

        def exit_error() -> typing.NoReturn:
            1 / 0

        with ClosingSocketPair() as (rd, wr):
            self.assertEqual(wr.send(b"data"), 4)

            _handle = evl.alarm(0.01, exit_clean)
            _handle = evl.alarm(0.005, say_hello)
            idle_handle = evl.enter_idle(say_waiting)
            if self._expected_idle_handle is not None:
                self.assertEqual(idle_handle, 1)

            evl.run()
            self.assertTrue("waiting" in out, out)
            self.assertTrue("hello" in out, out)
            self.assertTrue("clean exit" in out, out)
            handle = evl.watch_file(rd.fileno(), exit_clean)
            del out[:]

            evl.run()
            self.assertEqual(["clean exit"], out)
            self.assertTrue(evl.remove_watch_file(handle))
            _handle = evl.alarm(0, exit_error)
            self.assertRaises(ZeroDivisionError, evl.run)
            _handle = evl.watch_file(rd.fileno(), exit_error)
            self.assertRaises(ZeroDivisionError, evl.run)

    def test_run_in_executor(self):
        out: list[str] = []

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:

            def start():
                out.append("started")
                self.evl.run_in_executor(executor, func).add_done_callback(callback)

            def func() -> bool:
                out.append("future called")
                return True

            def callback(fut: Future) -> None:
                out.append(f"callback called with future outcome: {fut.result()}")

            def exit_clean() -> typing.NoReturn:
                out.append("clean exit")
                raise urwid.ExitMainLoop

            _handle = self.evl.alarm(0.1, exit_clean)
            _handle = self.evl.alarm(0.005, start)
            self.evl.run()

        self.assertEqual(
            [
                "started",
                "future called",
                "callback called with future outcome: True",
                "clean exit",
            ],
            out,
        )


class SelectEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        self.evl = urwid.SelectEventLoop()


@unittest.skipIf(IS_WINDOWS, "Windows is temporary not supported by AsyncioEventLoop.")
class AsyncioEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        if sys.version_info[:2] < (3, 11):
            self.loop = asyncio.get_event_loop_policy().get_event_loop()
            self.evl_runner = None
        else:
            self.evl_runner = asyncio.Runner(loop_factory=asyncio.SelectorEventLoop)
            self.loop = self.evl_runner.get_loop()
        self.evl = urwid.AsyncioEventLoop(loop=self.loop)

    def tearDown(self):
        if self.evl_runner is not None:
            self.evl_runner.close()

    _expected_idle_handle = None

    def test_error(self):
        evl = self.evl
        evl.alarm(0, lambda: 1 / 0)  # Simulate error in event loop
        self.assertRaises(ZeroDivisionError, evl.run)

    @unittest.skipIf(
        sys.implementation.name == "pypy",
        "Well known dead wait (lock?) on pypy.",
    )
    def test_coroutine_error(self):
        evl = self.evl

        async def error_coro():
            result = 1 / 0  # Simulate error in coroutine
            return result

        asyncio.ensure_future(error_coro(), loop=self.loop)
        self.assertRaises(ZeroDivisionError, evl.run)


@unittest.skip("Unstable test: race conditions happens")
@unittest.skipUnless(GLIB_AVAILABLE, "GLIB unavailable")
class GLibEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        self.evl = urwid.GLibEventLoop()

    def test_error(self):
        evl = self.evl
        evl.alarm(0, lambda: 1 / 0)  # Simulate error in event loop
        self.assertRaises(ZeroDivisionError, evl.run)


@unittest.skipUnless(TORNADO_AVAILABLE, "Tornado not available")
@unittest.skipIf(
    IS_WINDOWS,
    "Windows is temporary not supported by TornadoEventLoop due to race conditions.",
)
class TornadoEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        from tornado.ioloop import IOLoop

        self.evl = urwid.TornadoEventLoop(IOLoop())

    _expected_idle_handle = None


@unittest.skipUnless(TWISTED_AVAILABLE, "Twisted is not available")
@unittest.skipIf(
    IS_WINDOWS,
    "Windows is temporary not supported by TwistedEventLoop due to race conditions.",
)
class TwistedEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        self.evl = urwid.TwistedEventLoop()

    # can't restart twisted reactor, so use shortened tests
    def test_event_loop(self):
        pass

    def test_remove_alarm(self):
        pass

    def test_remove_watch_file(self):
        pass

    def test_run(self):
        from twisted.internet import threads

        evl = self.evl
        out = []
        wr: socket.socket
        rd: socket.socket

        def step2():
            out.append(rd.recv(2).decode("ascii"))

        def say_hello():
            out.append("hello")

        def say_waiting():
            out.append("waiting")

        def test_remove_alarm():
            handle = evl.alarm(50, lambda: None)
            self.assertTrue(evl.remove_alarm(handle))
            self.assertFalse(evl.remove_alarm(handle))
            out.append("remove_alarm ok")

        def test_remove_watch_file():
            with ClosingSocketPair() as (rd_, _wr):
                handle = evl.watch_file(rd_.fileno(), lambda: None)
                self.assertTrue(evl.remove_watch_file(handle))
                self.assertFalse(evl.remove_watch_file(handle))

            out.append("remove_watch_file ok")

        def test_threaded():
            out.append("schedule to thread")
            threads.deferToThread(func).addCallback(callback)

        def func() -> bool:
            out.append("future called")
            return True

        def callback(result: bool) -> None:
            out.append(f"callback called with future outcome: {result}")

        def exit_clean() -> typing.NoReturn:
            out.append("clean exit")
            raise urwid.ExitMainLoop

        def exit_error() -> typing.NoReturn:
            1 / 0

        with ClosingSocketPair() as (rd, wr):
            self.assertEqual(wr.send(b"data"), 4)

            _handle = evl.watch_file(rd.fileno(), step2)
            _handle = evl.alarm(0.1, exit_clean)
            _handle = evl.alarm(0.05, say_hello)
            _handle = evl.alarm(0.06, test_remove_alarm)
            _handle = evl.alarm(0.07, test_remove_watch_file)
            _handle = evl.alarm(0.08, test_threaded)
            self.assertEqual(evl.enter_idle(say_waiting), 1)
            evl.run()

        self.assertIn("da", out)
        self.assertIn("ta", out)
        self.assertIn("hello", out)
        self.assertIn("remove_alarm ok", out)
        self.assertIn("clean exit", out)
        self.assertIn("schedule to thread", out)
        self.assertIn("future called", out)
        self.assertIn("callback called with future outcome: True", out)

    def test_error(self):
        evl = self.evl
        evl.alarm(0, lambda: 1 / 0)  # Simulate error in event loop
        self.assertRaises(ZeroDivisionError, evl.run)

    @unittest.skip("not implemented for twisted event loop")
    def test_run_in_executor(self):
        """Not implemented for twisted event loop."""


@unittest.skipUnless(TRIO_AVAILABLE, "Trio not available")
@unittest.skipIf(
    IS_WINDOWS,
    "Windows is temporary not supported by TrioEventLoop due to race conditions.",
)
class TrioEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        self.evl = urwid.TrioEventLoop()

    _expected_idle_handle = None

    def test_error(self):
        evl = self.evl
        evl.alarm(0, lambda: 1 / 0)  # Simulate error in event loop
        self.assertRaises(ZeroDivisionError, evl.run)

    @unittest.skip("not implemented for trio event loop")
    def test_run_in_executor(self):
        """Not implemented for trio event loop."""


@unittest.skipUnless(ZMQ_AVAILABLE, "ZMQ is not available")
@unittest.skipIf(IS_WINDOWS, "ZMQEventLoop is not supported under windows")
class ZMQEventLoopTest(unittest.TestCase, EventLoopTestMixin):
    def setUp(self):
        self.evl = urwid.ZMQEventLoop()



from __future__ import annotations

import urwid


class CursorPudding(urwid.Widget):
    _sizing = frozenset(["flow"])
    _selectable = True

    def __init__(self):
        super().__init__()
        self.cursor_col = 0

    def rows(self, size: tuple[int], focus: bool = False) -> int:
        return 1

    def render(self, size: tuple[int], focus: bool = False) -> urwid.TextCanvas:
        (maxcol,) = size
        num_pudding = maxcol // len("Pudding")
        cursor = None
        if focus:
            cursor = self.get_cursor_coords(size)
        return urwid.TextCanvas([b"Pudding" * num_pudding], [], cursor=cursor, maxcol=maxcol)

    def get_cursor_coords(self, size: tuple[int]) -> tuple[int, int]:
        (maxcol,) = size
        col = min(self.cursor_col, maxcol - 1)
        return col, 0

    def keypress(self, size: tuple[int], key: str) -> str | None:
        (maxcol,) = size
        if key == "left":
            col = self.cursor_col - 1
        elif key == "right":
            col = self.cursor_col + 1
        else:
            return key
        self.cursor_x = max(0, min(maxcol - 1, col))
        self._invalidate()
        return None



# Urwid basic widget classes
#    Copyright (C) 2004-2012  Ian Ward
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Urwid web site: https://urwid.org/


from __future__ import annotations

import abc
import functools
import logging
import typing
import warnings
from operator import attrgetter

from urwid import signals
from urwid.canvas import Canvas, CanvasCache, CompositeCanvas
from urwid.command_map import command_map
from urwid.split_repr import split_repr
from urwid.util import MetaSuper

from .constants import Sizing

if typing.TYPE_CHECKING:
    from collections.abc import Callable, Hashable


__all__ = (
    "AbstractBoxWidget",
    "AbstractFixedWidget",
    "AbstractFlowWidget",
    "AbstractWidget",
    "Widget",
    "WidgetError",
    "WidgetMeta",
    "WidgetWarning",
    "WidgetWrap",
    "WidgetWrapError",
    "cache_widget_render",
    "cache_widget_rows",
    "delegate_to_widget_mixin",
    "fixed_size",
    "nocache_widget_render",
    "nocache_widget_render_instance",
)


WrappedWidget = typing.TypeVar("WrappedWidget", bound="AbstractWidget")
LOGGER = logging.getLogger(__name__)


class WidgetMeta(
    signals.MetaSignals,
    type(typing.Protocol),  # type: ignore[misc]  # hack: _ProtocolMeta is private, but need to subclass protocols
    MetaSuper,
):
    """
    Bases: :class:`MetaSuper`, :class:`MetaSignals`

    Automatic caching of render and rows methods.

    Class variable *no_cache* is a list of names of methods to not cache automatically.
    Valid method names for *no_cache* are ``'render'`` and ``'rows'``.

    Class variable *ignore_focus* if defined and set to ``True`` indicates
    that the canvas this widget renders is not affected by the focus parameter, so it may be ignored when caching.
    """

    def __init__(
        cls,
        name: str,
        bases: tuple[type, ...],
        d: dict[str, typing.Any],
    ) -> None:
        no_cache = d.get("no_cache", [])

        super().__init__(name, bases, d)

        if "render" in d:
            if "render" not in no_cache:
                render_fn = cache_widget_render(cls)
            else:
                render_fn = nocache_widget_render(cls)
            cls.render = render_fn

        if "rows" in d and "rows" not in no_cache:
            cls.rows = cache_widget_rows(cls)
        if "no_cache" in d:
            del cls.no_cache
        if "ignore_focus" in d:
            del cls.ignore_focus


@typing.runtime_checkable
class AbstractWidget(typing.Protocol):
    """Protocol for widget.

    Act as protocol for attributes-based type checking and as abstract base class for widget implementation.
    While normal widgets are expected to subclass :class:`Widget`,
    in some special cases we may deal with "implementation from scratch".
    In this case `isinstance` will fail to compare with :class:`Widget`,
    but if all required interfaces present - we can use it as valid widget implementation.

    .. note: sizing specific arguments left `typing.Any` to prevent static type checking errors.
    .. note: focus_position is not listed since it's raising `IndexError` on attribute access check
    """

    # Base widget methods (from Widget)
    @abc.abstractmethod
    def sizing(self) -> frozenset[Sizing]: ...

    @abc.abstractmethod
    def selectable(self) -> bool: ...

    @abc.abstractmethod
    def pack(
        self,
        size: typing.Any,
        focus: bool = False,
    ) -> tuple[int, int]: ...

    @property
    @abc.abstractmethod
    def base_widget(self) -> AbstractWidget: ...

    @property
    @abc.abstractmethod
    def focus(self) -> AbstractWidget | None: ...

    @abc.abstractmethod
    def keypress(self, size: typing.Any, key: str) -> str | None: ...

    @abc.abstractmethod
    def mouse_event(
        self,
        size: typing.Any,
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None: ...

    @abc.abstractmethod
    def render(
        self,
        size: typing.Any,
        focus: bool = False,
    ) -> Canvas: ...

    # Protected methods

    @abc.abstractmethod
    def _invalidate(self) -> None: ...


@typing.runtime_checkable
class AbstractBoxWidget(AbstractWidget, typing.Protocol):
    """Box widget protocol.

    Used internally to declare explicit required API.
    """

    @abc.abstractmethod
    def keypress(self, size: tuple[int, int], key: str) -> str | None: ...

    @abc.abstractmethod
    def mouse_event(
        self,
        size: tuple[int, int],
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None: ...

    @abc.abstractmethod
    def pack(
        self,
        size: tuple[int, int],
        focus: bool = False,
    ) -> tuple[int, int]: ...

    @abc.abstractmethod
    def render(
        self,
        size: tuple[int, int],
        focus: bool = False,
    ) -> Canvas: ...


@typing.runtime_checkable
class AbstractFlowWidget(AbstractWidget, typing.Protocol):
    """Flow widget protocol.

    Used internally to declare explicit required API.
    """

    @abc.abstractmethod
    def keypress(self, size: tuple[int], key: str) -> str | None: ...

    @abc.abstractmethod
    def mouse_event(
        self,
        size: tuple[int],
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None: ...

    @abc.abstractmethod
    def rows(self, size: tuple[int], focus: bool = False) -> int: ...

    @abc.abstractmethod
    def pack(
        self,
        size: tuple[int],
        focus: bool = False,
    ) -> tuple[int, int]: ...

    @abc.abstractmethod
    def render(
        self,
        size: tuple[int],
        focus: bool = False,
    ) -> Canvas: ...


@typing.runtime_checkable
class AbstractFixedWidget(AbstractWidget, typing.Protocol):
    """Fixed widget protocol.

    Used internally to declare explicit required API.
    """

    @abc.abstractmethod
    def keypress(self, size: tuple[()], key: str) -> str | None: ...

    @abc.abstractmethod
    def mouse_event(
        self,
        size: tuple[()],
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None: ...

    @abc.abstractmethod
    def pack(
        self,
        size: tuple[()],
        focus: bool = False,
    ) -> tuple[int, int]: ...

    @abc.abstractmethod
    def render(
        self,
        size: tuple[()],
        focus: bool = False,
    ) -> Canvas: ...


class WidgetError(Exception):
    """Widget specific errors."""


class WidgetWarning(Warning):
    """Widget specific warnings."""


def validate_size(
    widget: AbstractWidget,
    size: tuple[()] | tuple[int] | tuple[int, int],
    canv: Canvas,
) -> None:
    """
    Raise a WidgetError if a canv does not match size.
    """
    if (size and size[1:] != (0,) and size[0] != canv.cols()) or (len(size) > 1 and size[1] != canv.rows()):
        raise WidgetError(
            f"Widget {widget!r} rendered ({canv.cols():d} x {canv.rows():d}) canvas when passed size {size!r}!"
        )


def cache_widget_render(
    cls: WidgetMeta,
) -> Callable[[AbstractWidget, tuple[()] | tuple[int] | tuple[int, int], bool], Canvas]:
    """
    Return a function that wraps the cls.render() method and fetches and stores canvases with CanvasCache.
    """
    ignore_focus = bool(getattr(cls, "ignore_focus", False))
    fn = cls.render

    @functools.wraps(fn)
    def cached_render(
        self: AbstractWidget,
        size: tuple[()] | tuple[int] | tuple[int, int],
        focus: bool = False,
    ) -> Canvas:
        focus = focus and not ignore_focus

        if canv := CanvasCache.fetch(self, cls, size, focus):
            return canv

        canv = fn(self, size, focus=focus)  # type: ignore[call-arg]
        validate_size(self, size, canv)
        if canv.widget_info:
            canv = CompositeCanvas(canv)
        canv.finalize(self, size, focus)
        CanvasCache.store(cls, canv)
        return canv

    cached_render.original_fn = fn  # type: ignore[attr-defined]
    return cached_render


def nocache_widget_render(
    cls: WidgetMeta,
) -> Callable[[AbstractWidget, tuple[()] | tuple[int] | tuple[int, int], bool], Canvas]:
    """
    Return a function that wraps the cls.render() method
    and finalizes the canvas that it returns.
    """
    fn = cls.render
    if hasattr(fn, "original_fn"):
        fn = fn.original_fn

    @functools.wraps(fn)
    def finalize_render(
        self: AbstractWidget,
        size: tuple[()] | tuple[int] | tuple[int, int],
        focus: bool = False,
    ) -> Canvas:
        canv = fn(self, size, focus=focus)  # type: ignore[call-arg]
        if canv.widget_info:
            canv = CompositeCanvas(canv)
        validate_size(self, size, canv)
        canv.finalize(self, size, focus)
        return canv

    finalize_render.original_fn = fn  # type: ignore[attr-defined]  # un-cache
    return finalize_render


def nocache_widget_render_instance(
    self: AbstractWidget,
) -> Callable[[tuple[()] | tuple[int] | tuple[int, int], bool], Canvas]:
    """
    Return a function that wraps the cls.render() method
    and finalizes the canvas that it returns, but does not
    cache the canvas.
    """
    fn = self.render.original_fn  # type: ignore[attr-defined]

    @functools.wraps(fn)
    def finalize_render(
        size: tuple[()] | tuple[int] | tuple[int, int],
        focus: bool = False,
    ) -> Canvas:
        canv = fn(self, size, focus=focus)
        if canv.widget_info:
            canv = CompositeCanvas(canv)
        canv.finalize(self, size, focus)
        return canv

    finalize_render.original_fn = fn  # type: ignore[attr-defined]  # un-cache
    return finalize_render


def cache_widget_rows(cls: WidgetMeta) -> Callable[[AbstractWidget, tuple[int], bool], int]:
    """
    Return a function that wraps the cls.rows() method and returns rows from the CanvasCache if available.
    """
    ignore_focus = bool(getattr(cls, "ignore_focus", False))
    fn = cls.rows

    @functools.wraps(fn)
    def cached_rows(
        self: AbstractWidget,
        size: tuple[int],
        focus: bool = False,
    ) -> int:
        focus = focus and not ignore_focus

        if canv := CanvasCache.fetch(self, cls, size, focus):
            return canv.rows()

        return fn(self, size, focus)

    return cached_rows


class Widget(AbstractWidget, metaclass=WidgetMeta):
    """
    Widget base class

    .. attribute:: _selectable
       :annotation: = False

       The default :meth:`.selectable` method returns this value.

    .. attribute:: _sizing
       :annotation: = frozenset(['flow', 'box', 'fixed'])

       The default :meth:`.sizing` method returns this value.

    .. attribute:: _command_map
       :annotation: = urwid.command_map

       A shared :class:`CommandMap` instance. May be redefined in subclasses or widget instances.


    .. method:: rows(size, focus=False)

       .. note::

          This method is not implemented in :class:`.Widget` but
          must be implemented by any flow widget.  See :meth:`.sizing`.

       See :meth:`Widget.render` for parameter details.

       :returns: The number of rows required for this widget given a number of columns in *size*

       This is the method flow widgets use to communicate their size to other
       widgets without having to render a canvas. This should be a quick
       calculation as this function may be called a number of times in normal
       operation. If your implementation may take a long time you should add
       your own caching here.

       There is some metaclass magic defined in the :class:`Widget`
       metaclass :class:`WidgetMeta` that causes the
       result of this function to be retrieved from any
       canvas cached by :class:`CanvasCache`, so if your widget
       has been rendered you may not receive calls to this function. The class
       variable :attr:`ignore_focus` may be defined and set to ``True`` if this
       widget renders the same size regardless of the value of the *focus*
       parameter.

    .. method:: get_cursor_coords(size)

       .. note::

          This method is not implemented in :class:`.Widget` but
          must be implemented by any widget that may return cursor
          coordinates as part of the canvas that :meth:`render` returns.

       :param size: See :meth:`Widget.render` for details.
       :type size: widget size

       :returns: (*col*, *row*) if this widget has a cursor, ``None`` otherwise

       Return the cursor coordinates (*col*, *row*) of a cursor that will appear
       as part of the canvas rendered by this widget when in focus, or ``None``
       if no cursor is displayed.

       The :class:`ListBox` widget
       uses this method to make sure a cursor in the focus widget is not scrolled out of view.
       It is a separate method to avoid having to render the whole widget while calculating layout.

       Container widgets will typically call the :meth:`.get_cursor_coords` method on their focus widget.


    .. method:: get_pref_col(size)

       .. note::

          This method is not implemented in :class:`.Widget` but may be implemented by a subclass.

       :param size: See :meth:`Widget.render` for details.
       :type size: widget size

       :returns: a column number or ``'left'`` for the leftmost available
                 column or ``'right'`` for the rightmost available column

       Return the preferred column for the cursor to be displayed in this
       widget. This value might not be the same as the column returned from
       :meth:`get_cursor_coords`.

       The :class:`ListBox` and :class:`Pile`
       widgets call this method on a widget losing focus and use the value
       returned to call :meth:`.move_cursor_to_coords` on the widget becoming
       the focus. This allows the focus to move up and down through widgets
       while keeping the cursor in approximately the same column on screen.


    .. method:: move_cursor_to_coords(size, col, row)

       .. note::

          This method is not implemented in :class:`.Widget` but may be implemented by a subclass.
          Not implementing this method is equivalent to having a method that always returns
          ``False``.

       :param size: See :meth:`Widget.render` for details.
       :type size: widget size
       :param col: new column for the cursor, 0 is the left edge of this widget
       :type col: int
       :param row: new row for the cursor, 0 it the top row of this widget
       :type row: int

       :returns: ``True`` if the position was set successfully anywhere on *row*, ``False`` otherwise
    """

    _selectable = False
    _sizing = frozenset([Sizing.FLOW, Sizing.BOX, Sizing.FIXED])
    _command_map = command_map

    def __init__(self) -> None:
        self.logger = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

    def _invalidate(self) -> None:
        """Mark cached canvases rendered by this widget as dirty so that they will not be used again."""
        CanvasCache.invalidate(self)

    def _emit(self, name: Hashable, *args: typing.Any) -> None:
        """Convenience function to emit signals with self as first argument."""
        signals.emit_signal(self, name, self, *args)

    def selectable(self) -> bool:
        """
        :returns: ``True`` if this is a widget that is designed to take the
                  focus, i.e. it contains something the user might want to
                  interact with, ``False`` otherwise,

        This default implementation returns :attr:`._selectable`.
        Subclasses may leave these is if the are not selectable,
        or if they are always selectable they may
        set the :attr:`_selectable` class variable to ``True``.

        If this method returns ``True`` then the :meth:`.keypress` method
        must be implemented.

        Returning ``False`` does not guarantee that this widget will never be in
        focus, only that this widget will usually be skipped over when changing
        focus. It is still possible for non selectable widgets to have the focus
        (typically when there are no other selectable widgets visible).
        """
        return self._selectable

    def sizing(self) -> frozenset[Sizing]:
        """
        :returns: A frozenset including one or more of ``'box'``, ``'flow'`` and
                  ``'fixed'``.  Default implementation returns the value of
                  :attr:`._sizing`, which for this class includes all three.

        The sizing modes returned indicate the modes that may be
        supported by this widget, but is not sufficient to know
        that using that sizing mode will work.  Subclasses should
        make an effort to remove sizing modes they know will not
        work given the state of the widget, but many do not yet
        do this.

        If a sizing mode is missing from the set then the widget
        should fail when used in that mode.

        If ``'flow'`` is among the values returned then the other
        methods in this widget must be able to accept a
        single-element tuple (*maxcol*,) to their ``size``
        parameter, and the :meth:`rows` method must be defined.

        If ``'box'`` is among the values returned then the other
        methods must be able to accept a two-element tuple
        (*maxcol*, *maxrow*) to their size parameter.

        If ``'fixed'`` is among the values returned then the other
        methods must be able to accept an empty tuple () to
        their size parameter, and the :meth:`pack` method must
        be defined.
        """
        return self._sizing

    def pack(self, size: tuple[()] | tuple[int] | tuple[int, int], focus: bool = False) -> tuple[int, int]:
        """
        See :meth:`Widget.render` for parameter details.

        :returns: A "packed" size (*maxcol*, *maxrow*) for this widget

        Calculate and return a minimum
        size where all content could still be displayed. Fixed widgets must
        implement this method and return their size when ``()`` is passed as the
        *size* parameter.

        This default implementation returns the *size* passed, or the *maxcol*
        passed and the value of :meth:`rows` as the *maxrow* when (*maxcol*,)
        is passed as the *size* parameter.

        .. note::

           This is a new method that hasn't been fully implemented across the
           standard widget types. In particular it has not yet been
           implemented for container widgets.

        :class:`Text` widgets have implemented this method.
        You can use :meth:`Text.pack` to calculate the minimum
        columns and rows required to display a text widget without wrapping,
        or call it iteratively to calculate the minimum number of columns
        required to display the text wrapped into a target number of rows.
        """
        if not size:
            if Sizing.FIXED in self.sizing():
                raise NotImplementedError(f"{self!r} must override Widget.pack()")
            raise WidgetError(f"Cannot pack () size, this is not a fixed widget: {self!r}")

        if len(size) == 1:
            if Sizing.FLOW in self.sizing():
                # can not announce abstract
                return (*size, self.rows(size, focus))  # type: ignore[attr-defined]  # pylint: disable=no-member

            raise WidgetError(f"Cannot pack (maxcol,) size, this is not a flow widget: {self!r}")

        return size

    @property
    def base_widget(self) -> AbstractWidget:
        """Read-only property that steps through decoration widgets and returns the one at the base.

        This default implementation returns self.
        """
        return self

    @property
    def focus(self) -> AbstractWidget | None:
        """
        Read-only property returning the child widget in focus for container widgets.

        This default implementation always returns ``None``, indicating that this widget has no children.
        """
        return None

    @property
    def focus_position(self) -> typing.Any:
        """
        Property for reading and setting the focus position for container widgets.
        This default implementation raises :exc:`IndexError`,
        making normal widgets fail the same way accessing :attr:`.focus_position` on an empty container widget would.
        """
        raise IndexError(f"No focus_position, {self!r} is not a container widget")

    @focus_position.setter
    def focus_position(self, val: typing.Any) -> None:
        raise IndexError(f"No focus_position, {self!r} is not a container widget")

    def __repr__(self) -> str:
        """A friendly __repr__ for widgets.

        Designed to be extended by subclasses with _repr_words and _repr_attr methods.
        """
        return split_repr(self)

    def _repr_words(self) -> list[str]:
        words: list[str] = []
        if self.selectable():
            words = ["selectable", *words]
        if self.sizing() and self.sizing() != frozenset([Sizing.FLOW, Sizing.BOX, Sizing.FIXED]):
            words.append("/".join(sorted(self.sizing())))
        return [*words, "widget"]

    def _repr_attrs(self) -> dict[str, typing.Any]:
        return {}

    def keypress(
        self,
        size: tuple[()] | tuple[int] | tuple[int, int],
        key: str,
    ) -> str | None:
        """Keyboard input handler.

        :param size: See :meth:`Widget.render` for details
        :type size: tuple[()] | tuple[int] | tuple[int, int]
        :param key: a single keystroke value; see :ref:`keyboard-input`
        :type key: str
        :return: ``None`` if *key* was handled by *key* (the same value passed) if *key* was not handled
        :rtype: str | None
        """
        if not self.selectable():
            if hasattr(self, "logger"):
                self.logger.debug(f"keypress sent to non selectable widget {self!r}")
            else:
                warnings.warn(
                    f"Widget {self.__class__.__name__} did not call 'super().__init__()",
                    WidgetWarning,
                    stacklevel=3,
                )
                LOGGER.debug(f"Widget {self!r} is not selectable")
        return key

    def mouse_event(
        self,
        size: tuple[()] | tuple[int] | tuple[int, int],
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None:
        """Mouse event handler.

        :param size: See :meth:`Widget.render` for details.
        :type size: tuple[()] | tuple[int] | tuple[int, int]
        :param event: Values such as ``'mouse press'``, ``'ctrl mouse press'``,
                     ``'mouse release'``, ``'meta mouse release'``,
                     ``'mouse drag'``; see :ref:`mouse-input`
        :type event: str
        :param button: 1 through 5 for press events, often 0 for release events
                      (which button was released is often not known)
        :type button: int
        :param col: Column of the event, 0 is the left edge of this widget
        :type col: int
        :param row: Row of the event, 0 it the top row of this widget
        :type row: int
        :param focus: Set to ``True`` if this widget or one of its children is in focus
        :type focus: bool
        :return: ``True`` if the event was handled by this widget, ``False`` otherwise
        :rtype: bool | None
        """
        if not self.selectable():
            if hasattr(self, "logger"):
                self.logger.debug(f"Widget {self!r} is not selectable")
            else:
                warnings.warn(
                    f"Widget {self.__class__.__name__} not called 'super().__init__()",
                    WidgetWarning,
                    stacklevel=3,
                )
                LOGGER.debug(f"Widget {self!r} is not selectable")
        return False

    def render(
        self,
        size: tuple[()] | tuple[int] | tuple[int, int],
        focus: bool = False,
    ) -> Canvas:
        """Render widget and produce canvas

        :param size: One of the following, *maxcol* and *maxrow* are integers > 0:

            (*maxcol*, *maxrow*)
              for box sizing -- the parent chooses the exact
              size of this widget

            (*maxcol*,)
              for flow sizing -- the parent chooses only the
              number of columns for this widget

            ()
              for fixed sizing -- this widget is a fixed size
              which can't be adjusted by the parent
        :type size: widget size
        :param focus: set to ``True`` if this widget or one of its children is in focus
        :type focus: bool

        :returns: A :class:`Canvas` subclass instance containing the rendered content of this widget

        :class:`Text` widgets return a :class:`TextCanvas` (arbitrary text and display attributes),
        :class:`SolidFill` widgets return a :class:`SolidCanvas` (a single character repeated across the whole surface)
        and container widgets return a :class:`CompositeCanvas` (one or more other canvases arranged arbitrarily).

        If *focus* is ``False``, the returned canvas may not have a cursor position set.

        There is some metaclass magic defined in the :class:`Widget` metaclass :class:`WidgetMeta`
        that causes the result of this method to be cached by :class:`CanvasCache`.
        Later calls will automatically look up the value in the cache first.

        As a small optimization the class variable :attr:`ignore_focus`
        may be defined and set to ``True`` if this widget renders the same
        canvas regardless of the value of the *focus* parameter.

        Any time the content of a widget changes it should call
        :meth:`_invalidate` to remove any cached canvases, or the widget
        may render the cached canvas instead of creating a new one.
        """
        raise NotImplementedError


def fixed_size(size: tuple[()]) -> None:
    """
    raise ValueError if size != ().

    Used by FixedWidgets to test size parameter.
    """
    if size:
        raise ValueError(f"FixedWidget takes only () for size.passed: {size!r}")


def delegate_to_widget_mixin(attribute_name: str) -> type[Widget]:
    """
    Return a mixin class that delegates all standard widget methods
    to an attribute given by attribute_name.

    This mixin is designed to be used as a superclass of another widget.
    """
    # FIXME: this is so common, let's add proper support for it
    # when layout and rendering are separated

    get_delegate = attrgetter(attribute_name)

    class DelegateToWidgetMixin(Widget):
        no_cache: typing.ClassVar[list[str]] = ["rows"]  # crufty metaclass work-around

        def render(
            self,
            size: tuple[()] | tuple[int] | tuple[int, int],
            focus: bool = False,
        ) -> CompositeCanvas:
            canv = get_delegate(self).render(size, focus=focus)
            return CompositeCanvas(canv)

        @property
        def selectable(self) -> Callable[[], bool]:
            return get_delegate(self).selectable

        @property
        def get_cursor_coords(self) -> Callable[[tuple[()] | tuple[int] | tuple[int, int]], tuple[int, int] | None]:
            # TODO(Aleksei):  Get rid of property usage after getting rid of "if getattr"
            return get_delegate(self).get_cursor_coords

        @property
        def get_pref_col(self) -> Callable[[tuple[()] | tuple[int] | tuple[int, int]], int | None]:
            # TODO(Aleksei):  Get rid of property usage after getting rid of "if getattr"
            return get_delegate(self).get_pref_col

        def keypress(self, size: tuple[()] | tuple[int] | tuple[int, int], key: str) -> str | None:
            return get_delegate(self).keypress(size, key)

        @property
        def move_cursor_to_coords(self) -> Callable[[[tuple[()] | tuple[int] | tuple[int, int], int, int]], bool]:
            # TODO(Aleksei):  Get rid of property usage after getting rid of "if getattr"
            return get_delegate(self).move_cursor_to_coords

        @property
        def rows(self) -> Callable[[tuple[int], bool], int]:
            return get_delegate(self).rows

        @property
        def mouse_event(
            self,
        ) -> Callable[[tuple[()] | tuple[int] | tuple[int, int], str, int, int, int, bool], bool | None]:
            # TODO(Aleksei):  Get rid of property usage after getting rid of "if getattr"
            return get_delegate(self).mouse_event

        @property
        def sizing(self) -> Callable[[], frozenset[Sizing]]:
            return get_delegate(self).sizing

        @property
        def pack(self) -> Callable[[tuple[()] | tuple[int] | tuple[int, int], bool], tuple[int, int]]:
            return get_delegate(self).pack

    return DelegateToWidgetMixin


class WidgetWrapError(Exception):
    pass


class WidgetWrap(
    delegate_to_widget_mixin("_wrapped_widget"),  # type: ignore[misc]
    typing.Generic[WrappedWidget],
):
    def __init__(self, w: WrappedWidget) -> None:
        """
        w -- widget to wrap, stored as self._w

        This object will pass the functions defined in Widget interface
        definition to self._w.

        The purpose of this widget is to provide a base class for
        widgets that compose other widgets for their display and
        behaviour.  The details of that composition should not affect
        users of the subclass.  The subclass may decide to expose some
        of the wrapped widgets by behaving like a ContainerWidget or
        WidgetDecoration, or it may hide them from outside access.
        """
        super().__init__()
        if not isinstance(w, AbstractWidget):
            obj_class_path = f"{w.__class__.__module__}.{w.__class__.__name__}"
            warnings.warn(
                f"{obj_class_path} is not implementing Widget API",
                DeprecationWarning,
                stacklevel=2,
            )
        self._wrapped_widget = w

    @property
    def _w(self) -> WrappedWidget:
        return self._wrapped_widget

    @_w.setter
    def _w(self, new_widget: WrappedWidget) -> None:
        """
        Change the wrapped widget.  This is meant to be called
        only by subclasses.

        >>> size = (10,)
        >>> ww = WidgetWrap(Edit("hello? ", "hi"))
        >>> ww.render(size).text  # ... = b in Python 3
        [...'hello? hi ']
        >>> ww.selectable()
        True
        >>> ww._w = Text("goodbye")  # calls _set_w()
        >>> ww.render(size).text
        [...'goodbye   ']
        >>> ww.selectable()
        False
        """
        self._wrapped_widget = new_widget
        self._invalidate()

    def _set_w(self, w: WrappedWidget) -> None:
        """
        Change the wrapped widget.  This is meant to be called
        only by subclasses.
        >>> from urwid import Edit, Text
        >>> size = (10,)
        >>> ww = WidgetWrap(Edit("hello? ", "hi"))
        >>> ww.render(size).text  # ... = b in Python 3
        [...'hello? hi ']
        >>> ww.selectable()
        True
        >>> ww._w = Text("goodbye")  # calls _set_w()
        >>> ww.render(size).text
        [...'goodbye   ']
        >>> ww.selectable()
        False
        """
        warnings.warn(
            "_set_w is deprecated. Please use 'WidgetWrap._w' property directly. API will be removed in version 5.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._wrapped_widget = w
        self._invalidate()


def _test() -> None:
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    _test()



#!/usr/bin/python

"""Tests covering escape sequences processing"""

from __future__ import annotations

import unittest

from urwid.display import escape


class InputEscapeSequenceParserTest(unittest.TestCase):
    """Tests for parser of input escape sequences"""

    def test_bare_escape(self):
        codes = [27]
        expected = ["esc"]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_meta(self):
        codes = [27, ord("4"), ord("2")]
        expected = ["meta 4"]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([ord("2")], rest)

    def test_shift_arrows(self):
        codes = [27, ord("["), ord("a")]
        expected = ["shift up"]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_ctrl_pgup(self):
        codes = [27, 91, 53, 59, 53, 126]
        expected = ["ctrl page up"]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_esc_meta_1(self):
        codes = [27, 27, 49]
        expected = ["esc", "meta 1"]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_midsequence(self):
        # '[11~' is F1, '[12~' is F2, etc
        codes = [27, ord("["), ord("1")]

        with self.assertRaises(escape.MoreInputRequired):
            escape.process_keyqueue(codes, more_available=True)

        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(["meta ["], actual)
        self.assertListEqual([ord("1")], rest)

    def test_mouse_press(self):
        codes = [27, 91, 77, 32, 41, 48]
        expected = [("mouse press", 1.0, 8, 15)]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_bug_104(self):
        """GH #104: click-Esc & Esc-click crashes urwid apps"""
        codes = [27, 27, 91, 77, 32, 127, 59]
        expected = ["esc", ("mouse press", 1.0, 94, 26)]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

        codes = [27, 27, 91, 77, 35, 120, 59]
        expected = ["esc", ("mouse release", 0, 87, 26)]
        actual, rest = escape.process_keyqueue(codes, more_available=False)
        self.assertListEqual(expected, actual)
        self.assertListEqual([], rest)

    def test_functional_keys(self):
        """Test for functional keys F1-F4, F5-F12."""

        def check_key(expected: str, codes: list[int]) -> None:
            actual, rest = escape.process_keyqueue(codes, more_available=False)
            self.assertEqual(
                [expected],
                actual,
                f"Codes {codes!r} ({[chr(code) for code in codes]}) was not decoded to {expected!r}",
            )

        # F1-F4
        for num in range(1, 5):
            codes = [27, ord("O"), 79 + num]
            check_key(f"f{num}", codes)

        # F5-F8
        for num, offset in enumerate((0, 2, 3, 4), start=5):
            codes = [27, ord("["), ord("1"), 53 + offset, ord("~")]
            check_key(f"f{num}", codes)

        # F9-F12
        for num, offset in enumerate((0, 1, 3, 4), start=9):
            codes = [27, ord("["), ord("2"), 48 + offset, ord("~")]
            check_key(f"f{num}", codes)

    def test_functional_keys_mods_f1_f4(self):
        """Test for modifiers handling for functional keys F1-F4."""
        prefixes = ("O", "[1;")
        masks = "12345678"
        letters = "PQRS"

        for prefix in prefixes:
            encoded_prefix = tuple(ord(element) for element in prefix)
            for num, letter in enumerate(letters, start=1):
                for mask in masks:
                    codes = [27, *encoded_prefix, ord(mask), ord(letter)]
                    actual, rest = escape.process_keyqueue(codes, more_available=False)
                    expected = escape.escape_modifier(mask) + f"f{num}"
                    self.assertEqual(
                        [expected],
                        actual,
                        f"Codes {codes!r} ({[chr(code) for code in codes]}) was not decoded to {expected!r}",
                    )

    def test_functional_keys_mods_simple_f1_20(self):
        """Test for modifiers handling for functional keys F1-F20."""
        masks = "12345678"
        numbers = (11, 12, 13, 14, 15, 17, 18, 19, 20, 21, 23, 24, 25, 26, 28, 29, 31, 32, 33, 34)
        for num, number in enumerate(numbers, start=1):
            encoded_number = tuple(ord(element) for element in str(number))
            for mask in masks:
                codes = [27, ord("["), *encoded_number, ord(";"), ord(mask), ord("~")]
                actual, rest = escape.process_keyqueue(codes, more_available=False)
                expected = escape.escape_modifier(mask) + f"f{num}"
                self.assertEqual(
                    [expected],
                    actual,
                    f"Codes {codes!r} ({[chr(code) for code in codes]}) was not decoded to {expected!r}",
                )

    def test_sgrmouse(self):
        prefix = (27, ord("["), ord("<"))
        x = 4
        y = 8
        coord = (ord(f"{x + 1}"), ord(";"), ord(f"{y + 1}"))
        action = ord("M")
        modifiers = ((0, ""), (8, "meta "), (16, "ctrl "), (24, "meta ctrl "))
        for key, code in enumerate((0b0, 0b1, 0b10, 0b1000000, 0b1000001), start=1):
            for mod_code, mod in modifiers:
                key_code = tuple(ord(element) for element in str(code | mod_code))
                codes = [*prefix, *key_code, ord(";"), *coord, action]
                actual, rest = escape.process_keyqueue(codes, more_available=False)
                expected = [(mod + "mouse press", key, x, y)]
                self.assertEqual(
                    expected,
                    actual,
                    f"Codes {codes!r} ({[chr(code) for code in codes]}) was not decoded to {expected!r}",
                )



from __future__ import annotations

import unittest

import urwid
from urwid.util import get_encoding


class LineBoxTest(unittest.TestCase):
    def setUp(self) -> None:
        self.old_encoding = get_encoding()
        urwid.set_encoding("utf-8")

    def tearDown(self) -> None:
        urwid.set_encoding(self.old_encoding)

    def test_linebox_pack(self):
        # Bug #346 'pack' Padding does not run with LineBox
        t = urwid.Text("AAA\nCCC\nDDD")
        size = t.pack()
        l = urwid.LineBox(t)

        self.assertEqual(l.pack()[0], size[0] + 2)
        self.assertEqual(l.pack()[1], size[1] + 2)

    def test_border(self):
        wrapped = urwid.Text("Text\non\nfour\nlines", align=urwid.CENTER)
        l = urwid.LineBox(wrapped, tlcorner="╭", trcorner="╮", blcorner="╰", brcorner="╯")
        cols, rows = 7, 6
        self.assertEqual((cols, rows), l.pack(()))

        canvas = l.render(())
        self.assertEqual(cols, canvas.cols())
        self.assertEqual(rows, canvas.rows())

        self.assertEqual(
            [
                "╭─────╮",
                "│ Text│",
                "│  on │",
                "│ four│",
                "│lines│",
                "╰─────╯",
            ],
            [line.decode("utf-8") for line in canvas.text],
        )

    def test_header(self):
        wrapped = urwid.Text("Some text")
        l = urwid.LineBox(wrapped, title="Title", tlcorner="╭", trcorner="╮", blcorner="╰", brcorner="╯")
        cols, rows = 11, 3
        self.assertEqual((cols, rows), l.pack(()))

        canvas = l.render(())
        self.assertEqual(cols, canvas.cols())
        self.assertEqual(rows, canvas.rows())

        self.assertEqual(
            [
                "╭─ Title ─╮",
                "│Some text│",
                "╰─────────╯",
            ],
            [line.decode("utf-8") for line in canvas.text],
        )

        l.set_title("New")
        self.assertEqual(
            [
                "╭── New ──╮",
                "│Some text│",
                "╰─────────╯",
            ],
            [line.decode("utf-8") for line in l.render(()).text],
        )

    def test_negative(self):
        wrapped = urwid.Text("")
        with self.assertRaises(ValueError) as ctx:
            l = urwid.LineBox(wrapped, title="Title", tline="")

        self.assertEqual("Cannot have a title when tline is empty string", str(ctx.exception))

        l = urwid.LineBox(wrapped, tline="")
        with self.assertRaises(ValueError) as ctx:
            l.set_title("Fail")

        self.assertEqual("Cannot set title when tline is unset", str(ctx.exception))

    def test_partial_contour(self):
        def mark_pressed(btn: urwid.Button) -> None:
            nonlocal pressed

            pressed = True

        pressed = False
        wrapped = urwid.Button("Wrapped", on_press=mark_pressed)

        with self.subTest("No top line -> no top"):
            l = urwid.LineBox(wrapped, tline="")
            cols, rows = 13, 2
            self.assertEqual((cols, rows), l.pack(()))

            canvas = l.render(())
            self.assertEqual(cols, canvas.cols())
            self.assertEqual(rows, canvas.rows())

            self.assertEqual(
                [
                    "│< Wrapped >│",
                    "└───────────┘",
                ],
                [line.decode("utf-8") for line in canvas.text],
            )
            self.assertIsNone(l.keypress((), "enter"))
            self.assertTrue(pressed)

        pressed = False

        with self.subTest("No right side elements -> no side"):
            l = urwid.LineBox(wrapped, rline="")
            cols, rows = 12, 3
            self.assertEqual((cols, rows), l.pack(()))

            canvas = l.render(())
            self.assertEqual(cols, canvas.cols())
            self.assertEqual(rows, canvas.rows())

            self.assertEqual(
                [
                    "┌───────────",
                    "│< Wrapped >",
                    "└───────────",
                ],
                [line.decode("utf-8") for line in canvas.text],
            )
            self.assertIsNone(l.keypress((), "enter"))
            self.assertTrue(pressed)

        pressed = False

        with self.subTest("No left side elements -> no side"):
            l = urwid.LineBox(wrapped, lline="")
            cols, rows = 12, 3
            self.assertEqual((cols, rows), l.pack(()))

            canvas = l.render(())
            self.assertEqual(cols, canvas.cols())
            self.assertEqual(rows, canvas.rows())

            self.assertEqual(
                [
                    "───────────┐",
                    "< Wrapped >│",
                    "───────────┘",
                ],
                [line.decode("utf-8") for line in canvas.text],
            )
            self.assertIsNone(l.keypress((), "enter"))
            self.assertTrue(pressed)

        pressed = False

        with self.subTest("No bottom line -> no bottom"):
            l = urwid.LineBox(wrapped, bline="")
            cols, rows = 13, 2
            self.assertEqual((cols, rows), l.pack(()))

            canvas = l.render(())
            self.assertEqual(cols, canvas.cols())
            self.assertEqual(rows, canvas.rows())

            self.assertEqual(
                [
                    "┌───────────┐",
                    "│< Wrapped >│",
                ],
                [line.decode("utf-8") for line in canvas.text],
            )
            self.assertIsNone(l.keypress((), "enter"))
            self.assertTrue(pressed)

        pressed = False

    def test_columns_of_lineboxes(self):
        # BUG #748
        # Using PACK: width of widget 1 is 4, width of widget 2 is 5.
        # With equal weight widget 1 will be rendered also 5.
        columns = urwid.Columns(
            [
                (urwid.PACK, urwid.LineBox(urwid.Text("lol"), rline="", trcorner="", brcorner="")),
                (urwid.PACK, urwid.LineBox(urwid.Text("wtf"), tlcorner="┬", blcorner="┴")),
            ]
        )
        canvas = columns.render(())
        self.assertEqual(
            [
                "┌───┬───┐",
                "│lol│wtf│",
                "└───┴───┘",
            ],
            [line.decode("utf-8") for line in canvas.text],
        )

    def test_replace_widget(self):
        old_widget = urwid.Text("some text")
        lb_widget = urwid.LineBox(old_widget)
        canvas = lb_widget.render(())
        self.assertEqual(
            [
                "┌─────────┐",
                "│some text│",
                "└─────────┘",
            ],
            [line.decode("utf-8") for line in canvas.text],
        )
        self.assertIs(old_widget, lb_widget.original_widget)
        new_widget = urwid.widget.Button("now it's a button")
        lb_widget.original_widget = new_widget
        canvas = lb_widget.render(())
        self.assertEqual(
            [
                "┌─────────────────────┐",
                "│< now it's a button >│",
                "└─────────────────────┘",
            ],
            [line.decode("utf-8") for line in canvas.text],
        )
        self.assertIs(new_widget, lb_widget.original_widget)
